# Copyright (c) 2024 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import math
from contextlib import contextmanager
from functools import partial
from typing import Any, Callable, Dict, Optional, Tuple

import torch
from torch.nn.modules.loss import _Loss
from torch.utils.data import DataLoader

from nncf import NNCFConfig
from nncf.api.compression import CompressionAlgorithmController
from nncf.common.initialization.dataloader import NNCFDataLoader
from nncf.common.logging import nncf_logger
from nncf.common.logging.progress_bar import ProgressBar
from nncf.common.utils.api_marker import api
from nncf.config.structures import BNAdaptationInitArgs
from nncf.config.structures import ModelEvaluationArgs
from nncf.config.structures import QuantizationRangeInitArgs
from nncf.torch.nested_objects_traversal import objwalk
from nncf.torch.structures import AutoQPrecisionInitArgs
from nncf.torch.structures import DistributedCallbacksArgs
from nncf.torch.structures import ExecutionParameters
from nncf.torch.structures import LeGRInitArgs
from nncf.torch.structures import QuantizationPrecisionInitArgs
from nncf.torch.utils import default_distributed_unwrapper
from nncf.torch.utils import default_distributed_wrapper
from nncf.torch.utils import get_model_device
from nncf.torch.utils import is_tensor


class PTInitializingDataLoader(NNCFDataLoader):
    """
    This class wraps the torch.utils.data.DataLoader class,
    and defines methods to parse the general data loader output to
    separate the input to the compressed model and the ground truth target
    for the neural network. This is required for proper initialization of
    certain compression algorithms.
    """

    def __init__(self, data_loader: DataLoader):
        self._data_loader = data_loader

    @property
    def batch_size(self):
        return self._data_loader.batch_size

    def __iter__(self):
        return iter(self._data_loader)

    def __len__(self):
        return len(self._data_loader)

    def get_inputs(self, dataloader_output: Any) -> Tuple[Tuple, Dict]:
        """Returns (args, kwargs) for the current model call to be made during the initialization process"""
        raise NotImplementedError

    def get_target(self, dataloader_output: Any) -> Any:
        """
        Parses the generic data loader output and returns a structure to be used as
        ground truth in the loss criterion.

        :param dataloader_output - the (args, kwargs) tuple returned by the __next__ method.
        """

        raise NotImplementedError


class DefaultInitializingDataLoader(PTInitializingDataLoader):
    def get_inputs(self, dataloader_output: Any) -> Tuple[Tuple, Dict]:
        return (dataloader_output[0],), {}

    def get_target(self, dataloader_output: Any):
        return dataloader_output[1]


def wrap_dataloader_for_init(data_loader) -> PTInitializingDataLoader:
    if not isinstance(data_loader, PTInitializingDataLoader):
        loaded_item = next(iter(data_loader))
        if isinstance(loaded_item, (tuple, list)) and len(loaded_item) == 2:
            return DefaultInitializingDataLoader(data_loader)
        raise NotImplementedError(
            "Could not deduce the forward arguments from the initializing dataloader output.\n"
            "By default it is assumed that the data loader used for initialize "
            "produces a tuple/list of (*model_input*, *ground_truth*) and that no special "
            "forward arguments have to be set during init. If this is not the case, then instead "
            "of your regular data loader you need to pass a specialized version of "
            "PTInitializingDataLoader that returns a general (args, kwargs) tuple for your "
            "model to be called with at each __next__ call.\n"
            "See https://github.com/openvinotoolkit/nncf/blob/develop/docs/FAQ.md#pt_init_dataloader for "
            "an example of how to do this this in your code."
        )
    return data_loader


class PartialDataLoader:
    def __init__(self, regular_data_loader: DataLoader, iter_ratio=1.0):
        if iter_ratio < 0.0 or iter_ratio > 1.0:
            raise ValueError("iter_ratio must be within 0 to 1 range")
        self.data_loader = regular_data_loader
        self.batch_size = regular_data_loader.batch_size
        self._stop_id = math.ceil(len(self.data_loader) * iter_ratio)
        self._batch_id = 0

    def __iter__(self):
        self.data_loader_iter = iter(self.data_loader)
        self._batch_id = 0
        return self

    def __next__(self) -> Any:
        if self._batch_id < self._stop_id:
            loaded_item = next(self.data_loader_iter)
            self._batch_id += 1
            return loaded_item
        raise StopIteration

    def __len__(self) -> int:
        return self._stop_id


class DataLoaderBaseRunner:
    def __init__(self, model, init_device: Optional[str]):
        self.model = model
        self.init_device = init_device
        self.progressbar_description = "Algorithm initialization"

    def _run_model_inference(self, data_loader, num_init_steps, device):
        for i, loaded_item in ProgressBar(
            enumerate(data_loader),
            total=num_init_steps,
            desc=self.progressbar_description,
        ):
            if num_init_steps is not None and i >= num_init_steps:
                break
            args_kwargs_tuple = data_loader.get_inputs(loaded_item)
            self._infer_batch(args_kwargs_tuple, device)

    def _infer_batch(self, args_kwargs_tuple, device):
        to_device_fn = partial(torch.Tensor.to, device=device)
        args, kwargs = objwalk(args_kwargs_tuple, is_tensor, to_device_fn)
        self.model(*args, **kwargs)

    def run(self, data_loader, num_init_steps):
        if self.init_device is not None:
            original_device = get_model_device(self.model)
            self.model.to(self.init_device)

        self._prepare_initialization()
        device = get_model_device(self.model)
        data_loader = wrap_dataloader_for_init(data_loader)

        with torch.no_grad():
            self._run_model_inference(data_loader, num_init_steps, device)
            self._apply_initializers()

        if self.init_device is not None:
            self.model.to(original_device)

    def _prepare_initialization(self):
        raise NotImplementedError

    def _apply_initializers(self):
        raise NotImplementedError


class SimpleDataLoaderRunner(DataLoaderBaseRunner):
    def _prepare_initialization(self):
        pass

    def _apply_initializers(self):
        pass


class DataLoaderBNAdaptationRunner(DataLoaderBaseRunner):
    def __init__(self, model, init_device: str):
        super().__init__(model, init_device)
        self.progressbar_description = "BatchNorm statistics adaptation"
        self.original_momenta_values = {}
        self.original_training_state = {}

    @staticmethod
    def _apply_to_batchnorms(func):
        def func_apply_to_bns(module):
            if isinstance(
                module,
                (
                    torch.nn.modules.batchnorm.BatchNorm1d,
                    torch.nn.modules.batchnorm.BatchNorm2d,
                    torch.nn.modules.batchnorm.BatchNorm3d,
                ),
            ):
                func(module)

        return func_apply_to_bns

    @contextmanager
    def _bn_training_state_switcher(self) -> None:
        def save_original_bn_training_state(module: torch.nn.Module):
            self.original_training_state[module] = module.training

        def set_bn_training_state(module: torch.nn.Module, state: Dict[str, bool]):
            module.training = state

        def restore_original_bn_training_state(module: torch.nn.Module):
            module.training = self.original_training_state[module]

        self.model.apply(self._apply_to_batchnorms(save_original_bn_training_state))
        self.model.apply(self._apply_to_batchnorms(partial(set_bn_training_state, state=True)))
        try:
            yield
        finally:
            self.model.apply(self._apply_to_batchnorms(restore_original_bn_training_state))

    def _run_model_inference(self, data_loader, num_init_steps, device):
        with self._bn_training_state_switcher():
            for i, loaded_item in ProgressBar(
                enumerate(data_loader), total=num_init_steps, desc=self.progressbar_description
            ):
                if num_init_steps is not None and i >= num_init_steps:
                    break
                args_kwargs_tuple = data_loader.get_inputs(loaded_item)
                self._infer_batch(args_kwargs_tuple, device)

    def _prepare_initialization(self):
        pass

    def _apply_initializers(self):
        pass


def default_criterion_fn(outputs: Any, target: Any, criterion: Any) -> torch.Tensor:
    return criterion(outputs, target)


@api(canonical_alias="nncf.torch.register_default_init_args")
def register_default_init_args(
    nncf_config: NNCFConfig,
    train_loader: torch.utils.data.DataLoader,
    criterion: _Loss = None,
    criterion_fn: Callable[[Any, Any, _Loss], torch.Tensor] = None,
    train_steps_fn: Callable[
        [
            torch.utils.data.DataLoader,
            torch.nn.Module,
            torch.optim.Optimizer,
            CompressionAlgorithmController,
            Optional[int],
        ],
        type(None),
    ] = None,
    validate_fn: Callable[[torch.nn.Module, torch.utils.data.DataLoader], Tuple[float, float]] = None,
    val_loader: torch.utils.data.DataLoader = None,
    autoq_eval_fn: Callable[[torch.nn.Module, torch.utils.data.DataLoader], float] = None,
    model_eval_fn: Callable[[torch.nn.Module, torch.utils.data.DataLoader], float] = None,
    distributed_callbacks: Tuple[Callable, Callable] = None,
    execution_parameters: ExecutionParameters = None,
    legr_train_optimizer: torch.optim.Optimizer = None,
    device: str = None,
) -> NNCFConfig:
    nncf_config.register_extra_structs(
        [
            QuantizationRangeInitArgs(data_loader=wrap_dataloader_for_init(train_loader), device=device),
            BNAdaptationInitArgs(data_loader=wrap_dataloader_for_init(train_loader), device=device),
        ]
    )
    if train_loader and train_steps_fn and val_loader and validate_fn:
        nncf_config.register_extra_structs(
            [
                LeGRInitArgs(
                    train_loader=train_loader,
                    train_fn=train_steps_fn,
                    val_loader=val_loader,
                    val_fn=validate_fn,
                    train_optimizer=legr_train_optimizer,
                    nncf_config=nncf_config,
                )
            ]
        )

    if criterion is not None:
        if not criterion_fn:
            criterion_fn = default_criterion_fn
        nncf_config.register_extra_structs(
            [
                QuantizationPrecisionInitArgs(
                    criterion_fn=criterion_fn, criterion=criterion, data_loader=train_loader, device=device
                )
            ]
        )

    if autoq_eval_fn is not None:
        if not val_loader:
            val_loader = train_loader
        nncf_config.register_extra_structs(
            [AutoQPrecisionInitArgs(data_loader=val_loader, eval_fn=autoq_eval_fn, nncf_config=nncf_config)]
        )

    if model_eval_fn is not None:
        nncf_config.register_extra_structs([ModelEvaluationArgs(eval_fn=model_eval_fn)])

    if distributed_callbacks is None:
        distributed_callbacks = (
            partial(default_distributed_wrapper, execution_parameters=execution_parameters),
            default_distributed_unwrapper,
        )
    else:
        nncf_logger.info("Utilizing user-provided distributed training wrappers.")

    nncf_config.register_extra_structs([DistributedCallbacksArgs(*distributed_callbacks)])
    return nncf_config



# Copyright (c) 2024 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from collections import Counter
from typing import Any, Dict, Generator, List, Optional, Tuple

import networkx as nx
import networkx.algorithms.isomorphism as iso
import torch
from torch import Tensor

from nncf import nncf_logger
from nncf.common.graph.layer_attributes import BaseLayerAttributes
from nncf.common.graph.layer_attributes import Dtype
from nncf.torch.dynamic_graph.op_input_processing import OperatorInput
from nncf.torch.dynamic_graph.operation_address import OperationAddress
from nncf.torch.dynamic_graph.scope import Scope
from nncf.torch.dynamic_graph.trace_tensor import TensorMeta
from nncf.torch.dynamic_graph.trace_tensor import TracedParameter
from nncf.torch.dynamic_graph.trace_tensor import TracedTensor


class TensorMetaComparator:
    def __call__(self, lhs: TensorMeta, rhs: TensorMeta) -> bool:
        raise NotImplementedError


class DefaultTensorMetaComparator(TensorMetaComparator):
    def __call__(self, lhs: TensorMeta, rhs: TensorMeta) -> bool:
        return TensorMeta.default_comparator(lhs, rhs)


class ShapeIgnoringTensorMetaComparator(TensorMetaComparator):
    def __call__(self, lhs: TensorMeta, rhs: TensorMeta) -> bool:
        return lhs.creator_id == rhs.creator_id and lhs.index == rhs.index


class ShapeOnlyTensorMetaComparator(TensorMetaComparator):
    def __call__(self, lhs: TensorMeta, rhs: TensorMeta) -> bool:
        return lhs.shape[1:] == rhs.shape[1:]


class InputsMatcher:
    def __call__(
        self, node_inputs: List[TensorMeta], real_inputs: List[TensorMeta], tm_comparators: List[TensorMetaComparator]
    ) -> bool:
        raise NotImplementedError


class FirstInputsMatcher(InputsMatcher):
    def __call__(
        self, node_inputs: List[TensorMeta], real_inputs: List[TensorMeta], tm_comparators: List[TensorMetaComparator]
    ) -> bool:
        if not node_inputs or not real_inputs:
            return False

        if not node_inputs[0] or not real_inputs[0]:
            return False

        for tm_comparator in tm_comparators:
            if not tm_comparator(node_inputs[0], real_inputs[0]):
                return False
        return True


class DefaultInputsMatcher(InputsMatcher):
    def __call__(
        self,
        saved_inputs: List[TensorMeta],
        actual_inputs: List[TensorMeta],
        tm_comparators: List[TensorMetaComparator],
    ) -> bool:
        if saved_inputs is None and actual_inputs:
            return False

        matched_with_unexpected_tensors = False
        for saved_input, actual_input in zip(saved_inputs, actual_inputs):
            if saved_input is None and actual_input is None:
                continue
            if (saved_input is None) and (actual_input is not None):
                # torch.Tensor.size() seems to return ints when not tracing ONNX
                # and tensors when tracing ONNX. This breaks input-based node matching whenever
                # torch.Tensor.size() return value is passed into a NNCF-traced operation (such as `view`)
                # because at graph building time it expected to see ints as args and now it sees tensors.
                # To mitigate this, will only match inputs against the positions which had tensors during build-time
                # and disregard the rest of the argument positions.
                matched_with_unexpected_tensors = True
                continue
            if (saved_input is not None) and (actual_input is None):
                return False
            for tm_comparator in tm_comparators:
                if not tm_comparator(saved_input, actual_input):
                    return False
        if matched_with_unexpected_tensors:
            nncf_logger.debug(
                f"Had to match a node to an op which has tensors at positions where there were "
                f"no tensors at graph building time:\n"
                f"Node input metas: {saved_inputs}, but op input metas: {actual_inputs}"
            )
        return True


class OperationExecutionContext:
    """
    Information that allows to uniquely identify an operation inside the NNCF graph,
    i.e. determine whether an execution of the operator inside the module has already been
    registered as a node in the graph or not (in the latter case a new node would have to
    be created
    """

    def __init__(
        self,
        operator_name: str,
        scope_in_model: Scope,
        call_order: int,
        tensor_metas: List[TensorMeta],
        tm_comparators: List[TensorMetaComparator] = None,
        input_matcher: InputsMatcher = None,
    ):
        self.op_address = OperationAddress(operator_name, scope_in_model, call_order)
        # This should be a list with a length equal to the number of inputs.
        # "None" values in this list correspond to non-tensor input nodes.
        self.tensor_metas = tensor_metas
        self.tm_comparators = tm_comparators if tm_comparators else [DefaultTensorMetaComparator()]
        self.input_matcher = input_matcher if input_matcher else DefaultInputsMatcher()

    def __eq__(self, other):
        return self.op_address == other.op_address and Counter(self.tensor_metas) == Counter(other.tensor_metas)

    def matches_saved_inputs_from(self, other: "OperationExecutionContext"):
        # WARNING: not commutative
        return self.op_address == other.op_address and self.input_matcher(
            other.tensor_metas, self.tensor_metas, self.tm_comparators
        )

    def __hash__(self):
        return hash((self.operator_name, tuple(self.scope_in_model), self.call_order, tuple(self.tensor_metas)))

    def __str__(self):
        input_info_str = ""
        for meta in self.tensor_metas:
            if meta is None:
                input_info_str += "N;"
            else:
                input_info_str += str(meta) + ";"

        return super().__str__() + "(" + input_info_str + ")"

    @property
    def operator_name(self):
        return self.op_address.operator_name

    @property
    def scope_in_model(self) -> Scope:
        return self.op_address.scope_in_model

    @property
    def call_order(self):
        return self.op_address.call_order


class DynamicGraphNodeParameters:
    def __init__(
        self,
        layer_attributes: BaseLayerAttributes,
        ignored_algorithms: List[str],
        is_called_inside_nncf_module: bool,
        calling_module_id: int,
    ):
        self.layer_attributes = layer_attributes
        self.ignored_algorithms = ignored_algorithms
        self.is_called_inside_nncf_module = is_called_inside_nncf_module
        self.calling_module_id = calling_module_id


class DynamicGraphNode:
    def __init__(
        self,
        node_id: int,
        node_key: str,
        layer_attributes: BaseLayerAttributes,
        op_exec_context: OperationExecutionContext,
        calling_module_id: int,
        ignored_algorithms: List[str],
        is_called_inside_nncf_module: bool,
        is_in_iteration_scope: bool,
    ):
        self.node_id = node_id
        self.node_key = node_key
        self.layer_attributes = layer_attributes
        self.op_exec_context = op_exec_context
        self.calling_module_id = calling_module_id
        self.ignored_algorithms = ignored_algorithms
        self.is_called_inside_nncf_module = is_called_inside_nncf_module
        self.is_in_iteration_scope = is_in_iteration_scope

    @classmethod
    def build_from_nx_node(cls, nx_node: Dict[str, Any]) -> "DynamicGraphNode":
        return cls(
            node_id=nx_node[DynamicGraph.ID_NODE_ATTR],
            node_key=nx_node[DynamicGraph.KEY_NODE_ATTR],
            layer_attributes=nx_node.get(DynamicGraph.LAYER_ATTRIBUTES),
            op_exec_context=nx_node[DynamicGraph.OP_EXEC_CONTEXT_NODE_ATTR],
            ignored_algorithms=nx_node[DynamicGraph.IGNORED_ALGOS_NODE_ATTR],
            is_called_inside_nncf_module=nx_node[DynamicGraph.IS_CALLED_INSIDE_NNCF_MODULE],
            is_in_iteration_scope=nx_node[DynamicGraph.IS_IN_ITERATION_SCOPE_NODE_ATTR],
            calling_module_id=nx_node[DynamicGraph.CALLING_MODULE_ID],
        )

    def __eq__(self, other: "DynamicGraphNode") -> bool:
        return self.__dict__ == other.__dict__

    def __str__(self):
        return self.node_key


class DynamicGraphEdge:
    def __init__(
        self,
        from_node_id: int,
        to_node_id: int,
        activation_shape: List[int],
        input_port_id: int,
        output_port_id: int,
        dtype: Dtype,
        parallel_input_port_ids: List[int],
    ):
        """
        :param from_node_id - A numeric identifier of the starting node of the edge
        :param to_node_id - A numeric identifier of the node to which the edge is pointing
        :param activation_shape - The shape of the tensor associated with this edge
        :param input_port_id - Port of the operation consuming the tensor that this edge
          represents.
        :param output_port_id - Port of the operation that produced the tensor represented by this edge.
        :param dtype - Data type of the tensor represented by this edge.
        :param edge_multiplicity - Multiplicity of the edge. This is a workaround the the DiGraph's
          requirement to have no more than 1 edge between any two nodes, while, for instance, an operation
          of multiplying a tensor with itself, for instance, needs to somehow store in the graph
          the information that the operation actually has two inputs, even though one and the same
          tensor (i.e. edge) is used as both inputs; in this case `edge_multiplicity` must be set to 2.
        """
        self.from_node_id = from_node_id
        self.to_node_id = to_node_id
        self.activation_shape = activation_shape
        self.input_port_id = input_port_id
        self.output_port_id = output_port_id
        self.dtype = dtype
        self.parallel_input_port_ids = parallel_input_port_ids

    @classmethod
    def build_between_two_nx_nodes(
        cls, from_nx_node: Dict[str, Any], to_nx_node: Dict[str, Any], nx_edge: Dict[str, Any]
    ) -> "DynamicGraphEdge":
        from_node_id = from_nx_node[DynamicGraph.ID_NODE_ATTR]
        to_node_id = to_nx_node[DynamicGraph.ID_NODE_ATTR]
        return DynamicGraphEdge(
            from_node_id=from_node_id,
            to_node_id=to_node_id,
            activation_shape=nx_edge[DynamicGraph.ACTIVATION_SHAPE_EDGE_ATTR],
            input_port_id=nx_edge[DynamicGraph.INPUT_PORT_ID_EDGE_ATTR],
            output_port_id=nx_edge[DynamicGraph.OUTPUT_PORT_ID_EDGE_ATTR],
            dtype=nx_edge[DynamicGraph.ACTIVATION_DTYPE_EDGE_ATTR],
            parallel_input_port_ids=nx_edge[DynamicGraph.PARALLEL_INPUT_PORT_IDS],
        )


class DefaultScopeNodeMatcher:
    def __init__(self, node_id_to_key_dict, nx_graph):
        self._node_id_to_key_dict = node_id_to_key_dict
        self._nx_graph = nx_graph
        self._inputless_nodes: Dict[str, DynamicGraphNode] = {}

    def get_node_by_id(self, node_id):
        return self._nx_graph.nodes[self._node_id_to_key_dict[node_id]]

    def _find_nodes_with_matching_context_among_inputless(
        self, op_exec_context: OperationExecutionContext
    ) -> Dict[str, DynamicGraphNode]:
        node_candidates = {}
        for nx_node_key, node in self._inputless_nodes.items():
            if op_exec_context.matches_saved_inputs_from(node.op_exec_context):
                node_candidates[nx_node_key] = node
        return node_candidates

    def _find_nodes_with_matching_context_and_inputs(
        self, op_exec_context: OperationExecutionContext
    ) -> Dict[str, DynamicGraphNode]:
        nx_node_candidates = {}
        for info in op_exec_context.tensor_metas:
            if info is None or info.creator_id is None:
                continue
            creator_id = info.creator_id
            for successor_node_key in self._nx_graph.successors(self._node_id_to_key_dict[creator_id]):
                successor_node = self._nx_graph.nodes[successor_node_key]
                if op_exec_context.matches_saved_inputs_from(successor_node[DynamicGraph.OP_EXEC_CONTEXT_NODE_ATTR]):
                    nx_node_candidates[successor_node_key] = successor_node

        node_candidates: Dict[str, DynamicGraphNode] = {}
        for nx_node_key, nx_node_dict in nx_node_candidates.items():
            node_candidates[nx_node_key] = DynamicGraphNode.build_from_nx_node(nx_node_dict)

        return node_candidates

    def add_node(
        self,
        op_exec_context: OperationExecutionContext,
        inputs,
        node_parameters: DynamicGraphNodeParameters,
        is_in_iteration_scope: bool = False,
    ) -> DynamicGraphNode:
        node_id = len(self._node_id_to_key_dict)

        name_parts = (str(op_exec_context.scope_in_model), op_exec_context.operator_name)
        node_key = "{idx} {uri}".format(uri="/".join(name_parts), idx=node_id)

        nncf_logger.debug(f"New node added to NNCF graph: {node_key}")

        self._node_id_to_key_dict[node_id] = node_key
        attrs = {
            DynamicGraph.ID_NODE_ATTR: node_id,
            DynamicGraph.KEY_NODE_ATTR: node_key,
            DynamicGraph.OP_EXEC_CONTEXT_NODE_ATTR: op_exec_context,
            DynamicGraph.IS_IN_ITERATION_SCOPE_NODE_ATTR: is_in_iteration_scope,
            DynamicGraph.CALLING_MODULE_ID: node_parameters.calling_module_id,
        }
        if node_parameters.layer_attributes is not None:
            attrs[DynamicGraph.LAYER_ATTRIBUTES] = node_parameters.layer_attributes

        if node_parameters.ignored_algorithms is not None:
            attrs[DynamicGraph.IGNORED_ALGOS_NODE_ATTR] = node_parameters.ignored_algorithms
        else:
            attrs[DynamicGraph.IGNORED_ALGOS_NODE_ATTR] = []
        attrs[DynamicGraph.IS_CALLED_INSIDE_NNCF_MODULE] = node_parameters.is_called_inside_nncf_module

        self._nx_graph.add_node(node_key, **attrs)

        has_traced_inputs = False
        for i, info in enumerate(op_exec_context.tensor_metas):
            input_port_id = i
            if info is None or info.creator_id is None:
                continue

            has_traced_inputs = True
            parent = self._node_id_to_key_dict[info.creator_id]
            if self._nx_graph.get_edge_data(parent, node_key) is not None:
                self._nx_graph.edges[parent, node_key][DynamicGraph.PARALLEL_INPUT_PORT_IDS] += [input_port_id]
                continue
            self._nx_graph.add_edge(parent, node_key)
            self._nx_graph.edges[parent, node_key][DynamicGraph.ACTIVATION_SHAPE_EDGE_ATTR] = info.shape
            self._nx_graph.edges[parent, node_key][DynamicGraph.INPUT_PORT_ID_EDGE_ATTR] = input_port_id
            self._nx_graph.edges[parent, node_key][DynamicGraph.OUTPUT_PORT_ID_EDGE_ATTR] = info.index
            self._nx_graph.edges[parent, node_key][DynamicGraph.ACTIVATION_DTYPE_EDGE_ATTR] = info.dtype
            self._nx_graph.edges[parent, node_key][DynamicGraph.PARALLEL_INPUT_PORT_IDS] = []

        nx_node_dict = self._nx_graph.nodes[node_key]
        node = DynamicGraphNode.build_from_nx_node(nx_node_dict)

        if not has_traced_inputs:
            self._inputless_nodes[node_key] = node

        return node

    def find_node(
        self, op_address: OperationAddress, tensor_metas: List[TensorMeta], tm_comparators: List[TensorMetaComparator]
    ) -> DynamicGraphNode:
        op_exec_context = OperationExecutionContext(
            op_address.operator_name,
            op_address.scope_in_model,
            op_address.call_order,
            tensor_metas,
            tm_comparators=tm_comparators,
        )
        node_candidates = self._find_nodes_with_matching_context_and_inputs(op_exec_context)
        if not node_candidates:
            node_candidates = self._find_nodes_with_matching_context_among_inputless(op_exec_context)

        node_candidates = list(node_candidates.values())
        result = None
        if len(node_candidates) == 1:
            result = node_candidates[0]
        if len(node_candidates) > 1:
            nncf_logger.debug(f"More than one node was matched against context {op_exec_context}")
            result = node_candidates[0]

        return result


class IterationScopeNodeMatcher(DefaultScopeNodeMatcher):
    def __init__(self, node_id_to_key_dict, nx_graph):
        super().__init__(node_id_to_key_dict, nx_graph)
        self._first_iteration_nodes: {str: {str: DynamicGraphNode}} = {}

    def save_first_iteration_node(self, inputs: OperatorInput, node: DynamicGraphNode):
        """
        It finds and saves "starting" points of iteration for further matching with them on next iteration,
        instead of adding new nodes for each iteration. "Starting" points of iteration are nodes
            * that have at least one input node, which is outside of iteration scope
            * or whose all inputs are not tensors with TracedTensorMixin tracing capabilities.
        """
        op_exec_context = node.op_exec_context
        name = str(node)
        iter_scopes = op_exec_context.scope_in_model.get_iteration_scopes()
        if iter_scopes:
            for iter_scope in iter_scopes:
                if iter_scope not in self._first_iteration_nodes:
                    self._first_iteration_nodes[iter_scope] = {}
                first_nodes = self._first_iteration_nodes[iter_scope]
                has_input_outside_iteration = False
                untraced_tensor_inputs = []
                traced_tensor_inputs = []
                traced_parameter_inputs = []
                non_tensor_inputs = []
                for i in inputs:
                    input_obj = i.getter()
                    if isinstance(input_obj, Tensor):
                        if isinstance(input_obj, TracedTensor):
                            traced_tensor_inputs.append(input_obj)
                        elif isinstance(input_obj, TracedParameter):
                            traced_parameter_inputs.append(input_obj)
                        else:
                            untraced_tensor_inputs.append(input_obj)
                    else:
                        non_tensor_inputs.append(input_obj)

                for i in traced_parameter_inputs:
                    if i.tensor_meta is not None:
                        traced_tensor_inputs.append(i)

                for i in traced_tensor_inputs:
                    creator_id = i.tensor_meta.creator_id
                    creator_node = self.get_node_by_id(creator_id)
                    creator_node_op_exec_ctx = creator_node[DynamicGraph.OP_EXEC_CONTEXT_NODE_ATTR]
                    within_scopes = creator_node_op_exec_ctx.scope_in_model.get_iteration_scopes()
                    if iter_scope not in within_scopes:
                        has_input_outside_iteration = True

                if len(untraced_tensor_inputs) == (len(inputs) - len(non_tensor_inputs)):
                    has_input_outside_iteration = True
                if has_input_outside_iteration:
                    node_name = str(op_exec_context.op_address)
                    first_nodes[node_name] = node
                    nncf_logger.debug(f"Found first iteration node: {name} in scope: {iter_scope}")

    def add_node(
        self,
        op_exec_context: OperationExecutionContext,
        inputs,
        node_parameters: DynamicGraphNodeParameters,
        is_in_iteration_scope: bool = True,
    ) -> DynamicGraphNode:
        node = super().add_node(op_exec_context, inputs, node_parameters, is_in_iteration_scope=True)
        self.save_first_iteration_node(inputs, node)
        return node

    def find_node(
        self, op_address: OperationAddress, tensor_metas: List[TensorMeta], tm_comparators: List[TensorMetaComparator]
    ) -> Optional[DynamicGraphNode]:
        iter_scopes = op_address.scope_in_model.get_iteration_scopes()
        # compare meta information about first input nodes during the matching. During the iteration some nodes may
        # change number of inputs, e.g. on concat of hidden outputs
        input_matcher = FirstInputsMatcher()
        op_exec_context = OperationExecutionContext(
            op_address.operator_name,
            op_address.scope_in_model,
            op_address.call_order,
            tensor_metas,
            input_matcher=input_matcher,
            tm_comparators=tm_comparators,
        )
        node_candidates = self._find_nodes_with_matching_context_and_inputs(op_exec_context)
        if not node_candidates:
            op_exec_context = OperationExecutionContext(
                op_address.operator_name,
                op_address.scope_in_model,
                op_address.call_order,
                tensor_metas,
                tm_comparators=tm_comparators,
            )
            node_candidates = self._find_nodes_with_matching_context_among_inputless(op_exec_context)
            if not node_candidates and iter_scopes:
                # ignore information about node creator and index of input
                comparators = tm_comparators + [ShapeOnlyTensorMetaComparator()]
                op_exec_context = OperationExecutionContext(
                    op_address.operator_name,
                    op_address.scope_in_model,
                    op_address.call_order,
                    tensor_metas,
                    tm_comparators=comparators,
                )
                # match with starting points of iteration
                iter_nodes = self._match_first_iteration_nodes(op_exec_context, iter_scopes)
                for node_key, node in iter_nodes.items():
                    node_candidates[node_key] = node

        node_candidates = list(node_candidates.values())
        result = None
        if len(node_candidates) == 1:
            result = node_candidates[0]
        if len(node_candidates) > 1:
            nncf_logger.debug(f"More than one node was matched against context {op_exec_context}")
            result = node_candidates[0]

        return result

    def _match_first_iteration_nodes(self, op_exec_context: OperationExecutionContext, iter_scopes):
        node_candidates = {}
        for iter_scope in iter_scopes:
            if iter_scope in self._first_iteration_nodes:
                for name, node in self._first_iteration_nodes[iter_scope].items():
                    if op_exec_context.matches_saved_inputs_from(node.op_exec_context):
                        node_candidates[name] = node
                        break
                if node_candidates:
                    break
        return node_candidates

    def get_first_iteration_modules(self) -> Dict:
        return self._first_iteration_nodes


class NodeManager:
    def __init__(self, node_id_to_key_dict, nx_graph):
        self.base_matcher = DefaultScopeNodeMatcher(node_id_to_key_dict, nx_graph)
        self.iteration_matcher = IterationScopeNodeMatcher(node_id_to_key_dict, nx_graph)

    # TODO: optimize by matching exact module type
    @staticmethod
    def _within_iteration(scope: Scope):
        from nncf.torch.layers import ITERATION_MODULES

        for scope_element in scope.scope_elements:
            if scope_element.calling_module_class_name in ITERATION_MODULES.registry_dict:
                return True
        return False

    def choose_matcher(self, op_address: OperationAddress) -> DefaultScopeNodeMatcher:
        if self._within_iteration(op_address.scope_in_model):
            return self.iteration_matcher
        return self.base_matcher

    @staticmethod
    def choose_tm_comparators(
        op_address: OperationAddress, input_comparators_per_scope: List[Tuple[TensorMetaComparator, List[str]]]
    ) -> List[TensorMetaComparator]:
        result = []
        for pairs in input_comparators_per_scope:
            comparator, scopes = pairs
            for scope in scopes:
                if scope in str(op_address):
                    result.append(comparator)
        return result

    def find_node(
        self,
        op_address: OperationAddress,
        tensor_metas: List[TensorMeta],
        input_comparators_per_scope: List[Tuple[TensorMetaComparator, List[str]]],
    ) -> DynamicGraphNode:
        matcher = self.choose_matcher(op_address)
        comparators = self.choose_tm_comparators(op_address, input_comparators_per_scope)
        return matcher.find_node(op_address, tensor_metas, comparators)

    def add_node(
        self,
        op_address: OperationAddress,
        tensor_metas: List[TensorMeta],
        tm_comparators_per_scope: List[Tuple[TensorMetaComparator, List[str]]],
        inputs,
        node_parameters: DynamicGraphNodeParameters,
    ) -> DynamicGraphNode:
        matcher = self.choose_matcher(op_address)
        tm_comparators = self.choose_tm_comparators(op_address, tm_comparators_per_scope)
        op_exec_context = OperationExecutionContext(
            op_address.operator_name,
            op_address.scope_in_model,
            op_address.call_order,
            tensor_metas,
            tm_comparators=tm_comparators,
        )

        return matcher.add_node(op_exec_context, inputs, node_parameters)


class DynamicGraph:
    """
    The class for representing a graph dynamically built during a PyTorch model's `forward` method execution
    within a nncf.torch.dynamic_graph.context.TracingContext. This graph may change from a forward call to a
    forward call if the execution paths of the model change between the calls - this sets DynamicGraph apart from
    NNCFGraph which is a static representation of the model's structure. The DynamicGraph has limited support for
    RNN tracing and is rather suited to regular DNN tracing.
    """

    ID_NODE_ATTR = "id"
    KEY_NODE_ATTR = "key"
    LAYER_ATTRIBUTES = "layer_attributes"
    OP_EXEC_CONTEXT_NODE_ATTR = "op_exec_context"
    ACTIVATION_SHAPE_EDGE_ATTR = "activation_shape"
    ACTIVATION_DTYPE_EDGE_ATTR = "activation_dtype"
    INPUT_PORT_ID_EDGE_ATTR = "input_port_id"
    OUTPUT_PORT_ID_EDGE_ATTR = "output_port_id"
    IGNORED_ALGOS_NODE_ATTR = "ignored_algos"
    IS_CALLED_INSIDE_NNCF_MODULE = "is_called_inside_nncf_module"
    IS_IN_ITERATION_SCOPE_NODE_ATTR = "is_in_iteration_scope"
    CALLING_MODULE_ID = "calling_module_id"
    PARALLEL_INPUT_PORT_IDS = "parallel_input_port_ids"

    def __init__(self):
        self._nx_graph = nx.DiGraph()
        self._node_id_to_key_dict = {}
        self.match_manager = NodeManager(self._node_id_to_key_dict, self._nx_graph)
        self._input_nncf_nodes = []
        self._output_nncf_nodes = []
        self._integer_input_nodes = []

    def __eq__(self, other: "DynamicGraph"):
        nm = iso.categorical_node_match(
            [
                DynamicGraph.ID_NODE_ATTR,
                DynamicGraph.KEY_NODE_ATTR,
                DynamicGraph.OP_EXEC_CONTEXT_NODE_ATTR,
                DynamicGraph.LAYER_ATTRIBUTES,
                DynamicGraph.CALLING_MODULE_ID,
            ],
            [None, None, None, None, None],
        )
        em = iso.categorical_edge_match(
            [DynamicGraph.ACTIVATION_SHAPE_EDGE_ATTR, DynamicGraph.INPUT_PORT_ID_EDGE_ATTR], [None, None]
        )
        return nx.is_isomorphic(self._nx_graph, other._nx_graph, node_match=nm, edge_match=em)

    def find_node(
        self,
        op_address: OperationAddress,
        tensor_metas: List[TensorMeta],
        input_comparators_per_scope: List[Tuple[TensorMetaComparator, List[str]]],
    ) -> DynamicGraphNode:
        return self.match_manager.find_node(op_address, tensor_metas, input_comparators_per_scope)

    def add_node(
        self,
        op_address: OperationAddress,
        tensor_metas: List[TensorMeta],
        input_comparators_per_scope: List[Tuple[TensorMetaComparator, List[str]]],
        inputs: OperatorInput,
        node_parameters: DynamicGraphNodeParameters,
    ) -> DynamicGraphNode:
        node = self.match_manager.add_node(
            op_address, tensor_metas, input_comparators_per_scope, inputs, node_parameters
        )

        from nncf.common.graph.definitions import MODEL_INPUT_OP_NAME
        from nncf.common.graph.definitions import MODEL_OUTPUT_OP_NAME

        if node.op_exec_context.operator_name == MODEL_INPUT_OP_NAME:
            self._input_nncf_nodes.append(node)
            # Currently the MODEL_INPUT_OP_NAME node is added when an input is wrapped as
            # _ = nncf_model_input(input_tensor)
            # so it is expected that there 0-th positional arg will be the torch.Tensor we need to inspect
            tensor_input = inputs.op_args[0]
            assert isinstance(tensor_input, torch.Tensor)
            if tensor_input.dtype in (torch.int32, torch.int64, torch.long):
                self._integer_input_nodes.append(node)

        if node.op_exec_context.operator_name == MODEL_OUTPUT_OP_NAME:
            self._output_nncf_nodes.append(node)
        return node

    def get_input_nodes(self) -> List[DynamicGraphNode]:
        return self._input_nncf_nodes

    def is_integer_input_node(self, node: DynamicGraphNode) -> bool:
        return node in self._integer_input_nodes

    def get_output_nodes(self) -> List[DynamicGraphNode]:
        return self._output_nncf_nodes

    def get_nodes_count(self) -> int:
        return self._nx_graph.number_of_nodes()

    def get_all_nodes(self) -> List[DynamicGraphNode]:
        all_nodes = []
        for node_key in self._node_id_to_key_dict.values():
            dynamic_graph_node = DynamicGraphNode.build_from_nx_node(self._nx_graph.nodes[node_key])
            all_nodes.append(dynamic_graph_node)
        return all_nodes

    def get_all_edges(self) -> Generator[DynamicGraphEdge, None, None]:
        """
        Generates all edges in the graph
        """
        for from_nx_node_key, to_nx_node_key in self._nx_graph.in_edges:
            yield self._get_edge(self._get_node_by_key(from_nx_node_key), self._get_node_by_key(to_nx_node_key))

    def get_input_edges(self, node: DynamicGraphNode) -> List[DynamicGraphEdge]:
        """
        Returns edges of input tensors with description sorted by input port ID.

        :param node: Consumer node.
        :return: List of input edges for the node sorted by input port ID.
        """
        input_nodes = self._get_previous_nodes(node)
        edges = [self._get_edge(from_node, node) for from_node in input_nodes]
        return sorted(edges, key=lambda x: x.input_port_id)

    def set_layer_attributes_to_node(self, node: DynamicGraphNode, layer_attributes: BaseLayerAttributes):
        self._nx_graph.nodes[node.node_key][DynamicGraph.LAYER_ATTRIBUTES] = layer_attributes

    def is_graph_with_iteration_modules(self) -> bool:
        return len(self.match_manager.iteration_matcher.get_first_iteration_modules()) > 0

    def _get_node_by_key(self, key: str) -> DynamicGraphNode:
        """
        :param key: key (node_name) of the node.
        :return: DynamicGraphNode in a graph with such key.
        """
        return DynamicGraphNode.build_from_nx_node(self._nx_graph.nodes[key])

    def _get_previous_nodes(self, node: DynamicGraphNode) -> List[DynamicGraphNode]:
        nx_node_keys = self._nx_graph.pred[self._node_id_to_key_dict[node.node_id]]
        return [DynamicGraphNode.build_from_nx_node(self._nx_graph.nodes[key]) for key in nx_node_keys]

    def _get_edge(self, from_node: DynamicGraphNode, to_node: DynamicGraphNode) -> DynamicGraphEdge:
        nx_edge = self._get_nx_edge(from_node, to_node)
        from_nx_node = self._nx_graph.nodes[from_node.node_key]
        to_nx_node = self._nx_graph.nodes[to_node.node_key]
        return DynamicGraphEdge.build_between_two_nx_nodes(from_nx_node, to_nx_node, nx_edge)

    def _get_nx_edge(self, node_u: DynamicGraphNode, node_v: DynamicGraphNode):
        nx_node_u = self._nx_graph.nodes[self._node_id_to_key_dict[node_u.node_id]]
        nx_node_v = self._nx_graph.nodes[self._node_id_to_key_dict[node_v.node_id]]
        return self._nx_graph.edges[nx_node_u["key"], nx_node_v["key"]]



# Copyright (c) 2024 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import List

import torch

from nncf.experimental.torch.sparsity.movement.layers import MovementSparsifier
from nncf.torch.compression_method_api import PTCompressionLoss


class ImportanceLoss(PTCompressionLoss):
    """
    Module to calculate the compression loss of movement sparsity.
    """

    def __init__(self, operands: List[MovementSparsifier]):
        """
        Initializes the loss of movement sparsity in its algorithm controller.

        :param operands: List of movement sparsity operands for each layer to sparsify.
        """
        super().__init__()
        assert len(operands) > 0, "No sparse layers to calculate importance loss."
        self.operands = operands
        self._disabled = False

    def disable(self):
        self._disabled = True

    def calculate(self) -> torch.Tensor:
        loss = torch.tensor(0.0, device=self._get_device())
        if not self._disabled:
            for n, operand in enumerate(self.operands):
                loss = loss * (n / (n + 1)) + operand.loss() / (n + 1)  # avoid overflow
        return loss

    def _get_device(self) -> torch.device:
        return next(self.operands[0].parameters()).device



# Copyright (c) 2024 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Any

import torch


def clamp(x, low, high):
    return torch.max(torch.min(x, high), low)


def logit(x):
    return torch.log(x / (1 - x))


class STRound(torch.autograd.Function):
    @staticmethod
    def symbolic(g, input_, inplace=False):
        return g.op("STRound", input_)

    @staticmethod
    def forward(ctx, input_):
        output = input_.round()
        return output

    @staticmethod
    def backward(ctx: Any, *grad_outputs: Any) -> Any:
        return grad_outputs[0]


class STThreshold(torch.autograd.Function):
    @staticmethod
    def forward(ctx, input_, threshold: float = 0.5):
        output = (input_ > threshold).type(input_.dtype)
        return output

    @staticmethod
    def backward(ctx: Any, *grad_outputs: Any) -> Any:
        return grad_outputs[0], None



# Copyright (c) 2024 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from nncf.config.definitions import KNOWLEDGE_DISTILLATION_ALGO_NAME_IN_CONFIG
from nncf.config.definitions import ONLINE_DOCS_ROOT
from nncf.config.schemata.basic import NUMBER
from nncf.config.schemata.basic import STRING
from nncf.config.schemata.basic import with_attributes
from nncf.config.schemata.common.compression import BASIC_COMPRESSION_ALGO_SCHEMA
from nncf.config.schemata.defaults import KNOWLEDGE_DISTILLATION_SCALE
from nncf.config.schemata.defaults import KNOWLEDGE_DISTILLATION_TEMPERATURE

KNOWLEDGE_DISTILLATION_TYPE_OPTIONS = ["mse", "softmax"]

KNOWLEDGE_DISTILLATION_SCHEMA = {
    **BASIC_COMPRESSION_ALGO_SCHEMA,
    "description": f"This algorithm is only useful in combination with other compression algorithms and improves the"
    f"end accuracy result of the corresponding algorithm by calculating knowledge distillation loss "
    f"between the compressed model currently in training and its original, uncompressed counterpart. "
    f"See [KnowledgeDistillation.md]"
    f"({ONLINE_DOCS_ROOT}"
    f"/docs/compression_algorithms/KnowledgeDistillation.md) and the rest of this schema for "
    f"more details and parameters.",
    "properties": {
        "algorithm": {"const": KNOWLEDGE_DISTILLATION_ALGO_NAME_IN_CONFIG},
        "type": with_attributes(
            STRING, description="Type of Knowledge Distillation Loss.", enum=KNOWLEDGE_DISTILLATION_TYPE_OPTIONS
        ),
        "scale": with_attributes(
            NUMBER, description="Knowledge Distillation loss value multiplier", default=KNOWLEDGE_DISTILLATION_SCALE
        ),
        "temperature": with_attributes(
            NUMBER,
            description="`softmax` type only - Temperature for logits softening.",
            default=KNOWLEDGE_DISTILLATION_TEMPERATURE,
        ),
    },
    "additionalProperties": False,
    "required": ["type"],
}



# Copyright (c) 2024 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Dict, List

from nncf.common.graph import NNCFGraph
from nncf.common.graph import NNCFNode
from nncf.common.pruning.clusterization import Cluster
from nncf.common.pruning.clusterization import Clusterization
from nncf.common.pruning.operations import BasePruningOp
from nncf.common.pruning.utils import PruningAnalysisDecision
from nncf.common.pruning.utils import PruningAnalysisReason
from nncf.common.pruning.utils import PruningOperationsMetatypeRegistry
from nncf.common.pruning.utils import find_next_nodes_not_of_types
from nncf.common.pruning.utils import is_prunable_depthwise_conv


def get_position(nodes_list: List[NNCFNode], idx: int):
    for i, node in enumerate(nodes_list):
        if node.node_id == idx:
            return i
    return None


def merge_clusters_for_nodes(nodes_to_merge: List[NNCFNode], clusterization: Clusterization):
    """
    Merges clusters to which nodes from nodes_to_merge belongs.

    :param nodes_to_merge: All nodes are clusters for which should be merged.
    :param clusterization: Clusterization of nodes to work with.
    """
    if len(nodes_to_merge) <= 1:
        return

    # Will merge cluster with highest importance with others pairwise
    max_importance_node_id = None
    max_importance = 0
    for node in nodes_to_merge:
        importance = clusterization.get_cluster_containing_element(node.node_id).importance
        if importance > max_importance:
            max_importance_node_id = node.node_id
            max_importance = importance

    max_importance_cluster_id = clusterization.get_cluster_containing_element(max_importance_node_id).id
    for node in nodes_to_merge:
        if node.node_id != max_importance_node_id:
            current_node_cluster_id = clusterization.get_cluster_containing_element(node.node_id).id
            if current_node_cluster_id != max_importance_cluster_id:
                clusterization.merge_clusters(max_importance_cluster_id, current_node_cluster_id)


def cluster_special_ops(
    graph: NNCFGraph, special_types: List[str], identity_types: List[str]
) -> Clusterization[NNCFNode]:
    """
    This model will cluster all operations with type from special_types. Connected nodes is nodes that:
        1. Have path between nodes with only identity type nodes on it
        2. Have common input (identity type nodes can be on path from this input)

    :param graph: Graph to work with.
    :param special_types: List of types that should be grouped to groups of dependent nodes.
    :return: Clusterization of `special_types` nodes to the dependent groups.
    """
    topologically_sorted_nodes = graph.topological_sort()
    all_special_nodes = [node for node in graph.get_all_nodes() if node.node_type in special_types]

    # 0. Initially all nodes is a separate clusters
    clusterization = Clusterization[NNCFNode](lambda x: x.node_id)
    for i, node in enumerate(all_special_nodes):
        cluster = Cluster[NNCFNode](i, [node], [get_position(topologically_sorted_nodes, node.node_id)])
        clusterization.add_cluster(cluster)

    for node in topologically_sorted_nodes:
        if node.node_type in identity_types:
            continue

        all_outputs = find_next_nodes_not_of_types(graph, node, identity_types)
        all_output_special_nodes = [node for node in all_outputs if node.node_type in special_types]
        if node.node_type in special_types:
            all_output_special_nodes.append(node)
        merge_clusters_for_nodes(all_output_special_nodes, clusterization)

    return clusterization


class ModelAnalyzer:
    """
    Analyze the model before pruning to understand which parts could potentially be pruned without conflicts
    (all nodes that can't get pruned input will receive a non-pruned input).

    The algorithm consists of three steps:
        1. Set attribute `accept_pruned_input` to all nodes. This attribute shows whether this node can
        potentially get pruned input or node.
        2. Calculate `can_prune` attribute for all nodes by propagating `accept_pruned_input` up
        (from the result of the network to the inputs). Node can be pruned if all outputs of this node accept
        pruned input and all outputs can be pruned.
        3. Propagates `can_prune` down from input nodes to the outputs.

    As a result, all nodes are marked by the `can_prune` attribute as potentially prunable or not.
    """

    def __init__(
        self,
        graph: NNCFGraph,
        pruning_operator_metatypes: PruningOperationsMetatypeRegistry,
        prune_operations_types: List[str],
    ):
        """
        :param pruning_operator_metatypes: registry with operation metatypes pruning algorithm is aware of, i.e.
        metatypes describing operations with common pruning mask application and propagation properties, e.g.
        IdentityMaskForwardOps unifies operations that propagate pruning masks as is (relu, swish etc.), whereas
        Convolution unifies different convolution operations (conv1d, conv2d, conv3d) which accepts some input masks
        and provide some output masks.
        :param prune_operations_types: Types of operations with prunable parameters.
        """
        self.graph = graph
        self._pruning_operator_metatypes = pruning_operator_metatypes
        self._prune_operations_types = prune_operations_types
        pruning_op_metatypes_dict = self._pruning_operator_metatypes.registry_dict
        self._stop_propagation_op_metatype = pruning_op_metatypes_dict["stop_propagation_ops"]
        self._concat_op_metatype = pruning_op_metatypes_dict["concat"]

        self.can_prune = {idx: True for idx in self.graph.get_all_node_ids()}
        self.accept_pruned_input = {idx: True for idx in self.graph.get_all_node_ids()}

    def node_propagate_can_prune_attr(self, nncf_node: NNCFNode) -> bool:
        """
        Whether the node can propagate the `can_prune` attr. That means a node can propagate pruning mask
        (for example, activations propagate mask, but convolutions stop mask propagation).

        :param nncf_node: Node to work with.
        :return: Propagates this node can_prune throw or not.
        """
        node_type = nncf_node.node_type
        is_prunable = node_type in self._prune_operations_types
        return not is_prunable or (is_prunable and is_prunable_depthwise_conv(nncf_node))

    def node_accept_different_inputs(self, nncf_node: NNCFNode) -> bool:
        """
        Returns whether node accepts pruned and not pruned inputs as inputs at the same time.

        :return: Whether node accepts pruned and not pruned inputs as inputs at the same time.
        """
        return nncf_node.node_type in self._concat_op_metatype.get_all_op_aliases()

    def get_meta_operation_by_type_name(self, type_name: str) -> BasePruningOp:
        """
        Returns class of metaop that corresponds to `type_name` type.

        :return: Class of metaop that corresponds to `type_name` type.
        """
        cls = self._pruning_operator_metatypes.get_operator_metatype_by_op_name(type_name)
        if cls is None:
            cls = self._stop_propagation_op_metatype
        return cls

    def propagate_can_prune_attr_up(self):
        """
        Propagating can_prune attribute in reversed topological order.
        This attribute depends on accept_pruned_input and can_prune attributes of output nodes.
        Node can_prune is True if all outputs accept_pruned_input is True and all outputs
        (except convs because conv can be pruned by input and output independently).
        """
        reversed_sorted_nodes = reversed(self.graph.topological_sort())
        for node in reversed_sorted_nodes:
            # Check all output nodes accept_pruned_input attribute
            out_nodes = self.graph.get_next_nodes(node)
            outputs_accept_pruned_input = all(self.accept_pruned_input[node.node_id] for node in out_nodes)

            # Check all output nodes can_prune attribute
            outputs_will_be_pruned = all(
                self.can_prune[node.node_id] for node in out_nodes if self.node_propagate_can_prune_attr(node)
            )
            self.can_prune[node.node_id] = outputs_accept_pruned_input and outputs_will_be_pruned

    def propagate_can_prune_attr_down(self):
        """
        Propagating can_prune attribute down to fix all branching cases with one pruned and one not pruned
        branches.
        """
        sorted_nodes = self.graph.topological_sort()
        for node in sorted_nodes:
            # Propagate attribute only in not conv case
            if self.node_propagate_can_prune_attr(node):
                in_nodes = self.graph.get_previous_nodes(node)
                can_prune = all(self.can_prune[node.node_id] for node in in_nodes)
                can_prune_any = any(self.can_prune[node.node_id] for node in in_nodes)

                if (not self.node_accept_different_inputs(node) and not can_prune) or (
                    self.node_accept_different_inputs(node) and not can_prune_any
                ):
                    self.can_prune[node.node_id] = can_prune

    def set_accept_pruned_input_attr(self):
        for nncf_node in self.graph.get_all_nodes():
            cls = self.get_meta_operation_by_type_name(nncf_node.node_type)
            self.accept_pruned_input[nncf_node.node_id] = cls.accept_pruned_input(nncf_node)

    def analyse_model_before_pruning(self) -> Dict[int, PruningAnalysisDecision]:
        self.set_accept_pruned_input_attr()
        self.propagate_can_prune_attr_up()
        self.propagate_can_prune_attr_down()
        return {k: PruningAnalysisDecision(v, PruningAnalysisReason.MODEL_ANALYSIS) for k, v in self.can_prune.items()}



# Copyright (c) 2024 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import List

from nncf.common.graph.layer_attributes import BaseLayerAttributes


class TFNodeAttributes(BaseLayerAttributes):
    """
    Contains the TF-specific attributes of the NNCFNode.
    """

    def __init__(self, data_format: str):
        """
        Initializes the TF-specific attributes of the NNCFNode.

        :param data_format: The data format of the input and
            output data of the node. One of the following:
            `channels_last` or `channels_first`.
        """
        self._data_format = data_format

    def get_data_format(self) -> str:
        """
        Returns the data format of the input and output data of the node.

        :return: The data format of the input and output data of the node.
        """
        return self._data_format


class TFWeightedNodeAttributes(TFNodeAttributes):
    """
    Contains the TF-specific attributes of the NNCFNode with weight.
    """

    def __init__(self, data_format: str, weight_shape: List[int]):
        """
        Initializes the TF-specific attributes of the NNCFNode.

        :param data_format: The data format of the input and
            output data of the node. One of the following:
            `channels_last` or `channels_first`.
        :param weight_shape: The shape of the weight tensor.
        """
        super().__init__(data_format)
        self._weight_shape = weight_shape

    def get_weight_shape(self) -> List[int]:
        """
        Returns shape of the weight tensor.

        :return: Shape of the weight tensor.
        """
        return self._weight_shape



# Copyright (c) 2024 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Any, Dict, List

import torch

from nncf.torch.functions import logit
from nncf.torch.layer_utils import COMPRESSION_MODULES
from nncf.torch.layer_utils import CompressionParameter
from nncf.torch.layer_utils import StatefullModuleInterface
from nncf.torch.sparsity.layers import BinaryMask
from nncf.torch.sparsity.rb.functions import binary_mask
from nncf.torch.sparsity.rb.functions import calc_rb_binary_mask


@COMPRESSION_MODULES.register()
class RBSparsifyingWeight(BinaryMask, StatefullModuleInterface):
    WEIGHTS_SHAPE_KEY = "weight_shape"
    FROZEN_KEY = "frozen"
    COMPRESSION_LR_MULTIPLIER_KEY = "compression_lr_multiplier"
    EPS_KEY = "eps"

    def __init__(self, weight_shape: List[int], frozen=True, compression_lr_multiplier=None, eps=1e-6):
        super().__init__(weight_shape)
        self.frozen = frozen
        self.eps = eps
        self._mask = CompressionParameter(
            logit(torch.ones(weight_shape) * 0.99),
            requires_grad=not self.frozen,
            compression_lr_multiplier=compression_lr_multiplier,
        )
        self._compression_lr_multiplier = compression_lr_multiplier
        self.binary_mask = binary_mask(self._mask)
        self.register_buffer("uniform", torch.zeros(weight_shape))
        self.mask_calculation_hook = MaskCalculationHook(self)

    @property
    def mask(self) -> torch.nn.Parameter:
        return self._mask

    @mask.setter
    def mask(self, tensor: torch.Tensor):
        self._mask.data = tensor
        self.binary_mask = binary_mask(self._mask)

    def _calc_training_binary_mask(self, weight):
        u = self.uniform if self.training and not self.frozen else None
        return calc_rb_binary_mask(self._mask, u, self.eps)

    def loss(self):
        return binary_mask(self._mask)

    def get_config(self) -> Dict[str, Any]:
        return {
            self.WEIGHTS_SHAPE_KEY: list(self.mask.shape),
            self.FROZEN_KEY: self.frozen,
            self.COMPRESSION_LR_MULTIPLIER_KEY: self._compression_lr_multiplier,
            self.EPS_KEY: self.eps,
        }

    @classmethod
    def from_config(cls, state: Dict[str, Any]) -> "RBSparsifyingWeight":
        return RBSparsifyingWeight(
            weight_shape=state[cls.WEIGHTS_SHAPE_KEY],
            frozen=state[cls.FROZEN_KEY],
            compression_lr_multiplier=state[cls.COMPRESSION_LR_MULTIPLIER_KEY],
            eps=state[cls.EPS_KEY],
        )


class MaskCalculationHook:
    def __init__(self, module):
        self.hook = module._register_state_dict_hook(self.hook_fn)

    def hook_fn(self, module, destination, prefix, local_metadata):
        module.binary_mask = binary_mask(module.mask)
        destination[prefix + "_binary_mask"] = module.binary_mask
        return destination

    def close(self):
        self.hook.remove()



# Copyright (c) 2024 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import tensorflow as tf


def l1_filter_norm(weight_tensor, dim=0):
    """
    Calculates L1 for weight_tensor for the selected dimension.
    """
    transformed_tensor = _pull_tensor(weight_tensor, dim)
    return tf.norm(transformed_tensor, ord=1, axis=1)


def l2_filter_norm(weight_tensor, dim=0):
    """
    Calculates L2 for weight_tensor for the selected dimension.
    """
    transformed_tensor = _pull_tensor(weight_tensor, dim)
    return tf.norm(transformed_tensor, ord=2, axis=1)


def _pull_tensor(weight_tensor, dim=0):
    permutation = list(range(len(weight_tensor.shape)))
    permutation[0], permutation[dim] = permutation[dim], permutation[0]
    weight_tensor = tf.transpose(weight_tensor, perm=permutation)
    return tf.reshape(weight_tensor, [weight_tensor.shape[0], -1])


def tensor_l2_normalizer(weight_tensor):
    return tf.math.l2_normalize(weight_tensor)


def geometric_median_filter_norm(weight_tensor, dim=0):
    """
    Compute geometric median norm for filters.
    :param weight_tensor: tensor with weights
    :param dim: dimension of output channel
    :return: metric value for every weight from weights_tensor
    """
    weight_vec = _pull_tensor(weight_tensor, dim)
    square_norms = tf.reduce_sum(tf.square(weight_vec), axis=1, keepdims=True)
    similar_matrix = tf.sqrt(
        tf.maximum(
            square_norms - 2 * tf.matmul(weight_vec, weight_vec, transpose_b=True) + tf.transpose(square_norms), 0
        )
    )
    similar_sum = tf.reduce_sum(similar_matrix, axis=0)
    return similar_sum


def _l2_distance(x, y):
    return tf.sqrt(tf.reduce_sum(tf.square(x - y)))


FILTER_IMPORTANCE_FUNCTIONS = {
    "L2": l2_filter_norm,
    "L1": l1_filter_norm,
    "geometric_median": geometric_median_filter_norm,
}


def calculate_binary_mask(weight_importance, threshold):
    return tf.cast(weight_importance >= threshold, tf.float32)



# Copyright (c) 2024 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from nncf.api.compression import CompressionAlgorithmController
from nncf.common.composite_compression import CompositeCompressionAlgorithmController
from nncf.common.utils.api_marker import api
from nncf.tensorflow.pruning.base_algorithm import BasePruningAlgoController
from nncf.tensorflow.pruning.callbacks import PruningStatisticsCallback
from nncf.tensorflow.sparsity.base_algorithm import BaseSparsityController
from nncf.tensorflow.sparsity.callbacks import SparsityStatisticsCallback
from nncf.tensorflow.sparsity.callbacks import UpdateMask


@api(canonical_alias="nncf.tensorflow.create_compression_callbacks")
def create_compression_callbacks(
    compression_ctrl: CompressionAlgorithmController,
    log_tensorboard: bool = True,
    log_text: bool = True,
    log_dir: bool = None,
):
    compression_controllers = (
        compression_ctrl.child_ctrls
        if isinstance(compression_ctrl, CompositeCompressionAlgorithmController)
        else [compression_ctrl]
    )
    for ctrl in compression_controllers:
        if isinstance(ctrl, (BaseSparsityController, BasePruningAlgoController)):
            callbacks = [UpdateMask(ctrl.scheduler)]
            if log_tensorboard or log_text:
                if isinstance(ctrl, BaseSparsityController):
                    statistics_callback_cls = SparsityStatisticsCallback
                else:
                    statistics_callback_cls = PruningStatisticsCallback

                callbacks += [
                    statistics_callback_cls(
                        ctrl.statistics, log_tensorboard=log_tensorboard, log_text=log_text, log_dir=log_dir
                    )
                ]
            return callbacks
    return []



# Copyright (c) 2024 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import io
import os
import os.path as osp
import pathlib
from abc import ABC
from abc import abstractmethod
from typing import Any, Callable, Dict, List, Optional, Tuple, TypeVar, Union, cast

from nncf.api.compression import CompressionAlgorithmController
from nncf.api.compression import CompressionStage
from nncf.common.logging import nncf_logger
from nncf.common.plotting import noninteractive_plotting
from nncf.common.statistics import NNCFStatistics
from nncf.common.utils.helpers import configure_accuracy_aware_paths
from nncf.common.utils.tensorboard import prepare_for_tensorboard
from nncf.config.schemata.defaults import AA_COMPRESSION_RATE_STEP_REDUCTION_FACTOR
from nncf.config.schemata.defaults import AA_INITIAL_COMPRESSION_RATE_STEP
from nncf.config.schemata.defaults import AA_INITIAL_TRAINING_PHASE_EPOCHS
from nncf.config.schemata.defaults import AA_LR_REDUCTION_FACTOR
from nncf.config.schemata.defaults import AA_MAXIMAL_TOTAL_EPOCHS
from nncf.config.schemata.defaults import AA_MINIMAL_COMPRESSION_RATE_STEP
from nncf.config.schemata.defaults import AA_PATIENCE_EPOCHS

TModel = TypeVar("TModel")
OptimizerType = TypeVar("OptimizerType")
LRSchedulerType = TypeVar("LRSchedulerType")
TensorboardWriterType = TypeVar("TensorboardWriterType")
Checkpoint = TypeVar("Checkpoint")


Image = Any  # Default type for Image.

try:
    import matplotlib.pyplot as plt  # type: ignore
    import PIL.Image
    from PIL.Image import Image

    IMG_PACKAGES_AVAILABLE = True
except ImportError:
    IMG_PACKAGES_AVAILABLE = False

ImageType = TypeVar("ImageType", bound=Image)


class TrainingRunner(ABC):
    """
    Runner is an object that is used by a TrainingLoop instance to control the training process
    via wrapping user-supplied functions such as `train_epoch_fn` and `validate_fn`.
    """

    @abstractmethod
    def train_epoch(self, model: TModel, compression_controller: CompressionAlgorithmController) -> None:
        """
        Calls train_epoch_fn and compression_controller.scheduler.epoch_step()

        :param model: The model to be fine-tuned
        :param compression_controller: The compression controller to be used during
        model fine-tuning
        """

    @abstractmethod
    def validate(self, model: TModel) -> float:
        """
        Compute the target metric value on the validation dataset for the supplied model.

        :param model: The model to be evaluated
        :return: Target validation metric value (float).
        """

    @abstractmethod
    def dump_statistics(self, model: TModel, compression_controller: CompressionAlgorithmController) -> None:
        """
        Dumps current statistics from compression_controller and dumps model's checkpoint.

        :param model: The model
        :param compression_controller: The compression controller to be used during
        model fine-tuning
        """

    @abstractmethod
    def dump_checkpoint(self, model: TModel, compression_controller: CompressionAlgorithmController) -> None:
        """
        Dump current model checkpoint on disk.

        :param model: The model to be saved
        :param compression_controller: The compression controller to be used during
        model fine-tuning
        """

    @abstractmethod
    def configure_optimizers(self) -> None:
        """
        Initialize the training optimizer object (and, optionally, the learning rate scheduler object).
        """

    @abstractmethod
    def update_learning_rate(self) -> None:
        """
        Update learning rate.
        """

    @abstractmethod
    def reset_training(self) -> None:
        """
        Initialize all-training related parameters (e.g. epoch count, optimizer, learning rate scheduler).
        """

    @abstractmethod
    def calculate_minimal_tolerable_accuracy(self, uncompressed_model_accuracy: float) -> None:
        """
        :param uncompressed_model_accuracy: The uncompressed model accuracy.

        Calculate the minimal tolerable accuracy from thr uncompressed_model_accuracy and
        set the obtained value to the `minimal_tolerable_accuracy` attribute of the TrainingRunner.
        """

    @abstractmethod
    def initialize_training_loop_fns(
        self,
        train_epoch_fn: Callable[
            [
                CompressionAlgorithmController,
                TModel,
                Optional[int],
                Optional[OptimizerType],
                Optional[LRSchedulerType],
            ],
            float,
        ],
        validate_fn: Callable[[TModel], float],
        configure_optimizers_fn: Optional[Callable[[], Tuple[OptimizerType, LRSchedulerType]]],
        dump_checkpoint_fn: Optional[
            Callable[
                [
                    TModel,
                    CompressionAlgorithmController,
                    "TrainingRunner",
                    Optional[Union[str, pathlib.Path]],
                ],
                None,
            ]
        ],
        load_checkpoint_fn: Optional[Callable[[Checkpoint, Union[str, pathlib.Path]], None]] = None,
        early_stopping_fn: Optional[
            Callable[
                [
                    float,
                ],
                bool,
            ]
        ] = None,
        update_learning_rate_fn: Optional[Callable[[], None]] = None,
    ) -> None:
        """
        Register the user-supplied functions to be used to control the training process.

        :param train_epoch_fn: a method to fine-tune the model for a single epoch
        (to be called inside the `train_epoch` of the TrainingRunner).
        :param validate_fn: a method to evaluate the model on the validation dataset
        (to be called inside the `train_epoch` of the TrainingRunner).
        :param configure_optimizers_fn: a method to instantiate an optimizer and a learning
        rate scheduler (to be called inside the `configure_optimizers` of the TrainingRunner).
        :param dump_checkpoint_fn: a method to dump a checkpoint.
        """

    @abstractmethod
    def initialize_logging(
        self,
        log_dir: Optional[Union[str, pathlib.Path]] = None,
        tensorboard_writer: Optional[TensorboardWriterType] = None,
    ) -> None:
        """
        Initialize logging related variables

        :param log_dir: The path to be used for logging and checkpoint saving.
        :param tensorboard_writer: The tensorboard object to be used for logging.
        """

    @abstractmethod
    def load_best_checkpoint(self, model: TModel) -> float:
        """
        Load the most accurate model state from the fine-tuning history.

        param model: The model object in which the state will be loaded.
        return: Compression rate of the returned model.
        """

    @abstractmethod
    def is_model_fully_compressed(self, compression_controller: CompressionAlgorithmController) -> bool:
        """
        Check if model is fully compressed

        param compression_controller: Compression controller used to compress the model
        return: True if the model is fully compressed, and False otherwise
        """


class BaseAccuracyAwareTrainingRunner(TrainingRunner):
    """
    The base accuracy-aware training Runner object,
    initialized with the default parameters unless specified in the config.
    """

    def __init__(
        self,
        accuracy_aware_training_params: Dict[str, Union[float, int]],
        uncompressed_model_accuracy: float,
        verbose: bool = True,
        dump_checkpoints: bool = True,
        lr_updates_needed: bool = True,
    ):
        self.uncompressed_model_accuracy = uncompressed_model_accuracy
        self.maximal_relative_accuracy_drop: float = accuracy_aware_training_params.get(
            "maximal_relative_accuracy_degradation", 1.0
        )
        self.maximal_absolute_accuracy_drop = accuracy_aware_training_params.get(
            "maximal_absolute_accuracy_degradation"
        )
        self.maximal_total_epochs = accuracy_aware_training_params.get("maximal_total_epochs", AA_MAXIMAL_TOTAL_EPOCHS)

        self.verbose: bool = verbose
        self.dump_checkpoints: bool = dump_checkpoints

        self.base_lr_reduction_factor_during_search = 1.0
        self.lr_updates_needed: bool = lr_updates_needed

        self.accuracy_budget: Optional[float] = None
        self.is_higher_metric_better: bool = True
        self.optimizer: OptimizerType  # type: ignore
        self.lr_scheduler: LRSchedulerType  # type: ignore

        self.training_epoch_count: int = 0
        self.cumulative_epoch_count: int = 0
        self.best_val_metric_value: float = 0
        self.current_val_metric_value: float = 0
        self.current_loss: float = 0

        self._compressed_training_history: List[Tuple[float, float]] = []
        self._best_checkpoint: Optional[Tuple[Union[str, pathlib.Path], float]] = None

        self._train_epoch_fn: Callable[
            [
                CompressionAlgorithmController,
                TModel,
                Optional[int],
                Optional[OptimizerType],
                Optional[LRSchedulerType],
            ],
            float,
        ]
        self._validate_fn: Callable[[TModel], float]
        self._configure_optimizers_fn: Callable[[], Tuple[OptimizerType, LRSchedulerType]]
        self._dump_checkpoint_fn: Optional[
            Callable[
                [
                    TModel,
                    CompressionAlgorithmController,
                    "TrainingRunner",
                    Optional[Union[str, pathlib.Path]],
                ],
                None,
            ]
        ]
        self._load_checkpoint_fn: Optional[Callable[[Checkpoint, Union[str, pathlib.Path]], None]] = None
        self._early_stopping_fn: Optional[Callable[[float], bool]] = None
        self._update_learning_rate_fn: Optional[Callable[[], None]] = None

        self._log_dir: Union[str, pathlib.Path]
        self._checkpoint_save_dir: Union[str, pathlib.Path]
        self._tensorboard_writer: TensorboardWriterType  # type: ignore

    def train_epoch(self, model: TModel, compression_controller: CompressionAlgorithmController) -> None:
        compression_controller.scheduler.epoch_step()
        # assuming that epoch number is only used for logging in train_fn:
        self.current_loss = self._train_epoch_fn(  # type: ignore
            compression_controller,
            model,
            epoch=self.cumulative_epoch_count,
            optimizer=self.optimizer,
            lr_scheduler=self.lr_scheduler,
        )
        self.training_epoch_count += 1
        self.cumulative_epoch_count += 1

    def dump_statistics(self, model: TModel, compression_controller: CompressionAlgorithmController) -> None:
        statistics = cast(NNCFStatistics, compression_controller.statistics())

        if self.verbose:
            nncf_logger.info(statistics.to_str())

        self.add_tensorboard_scalar(
            "val/accuracy_aware/metric_value",
            self.current_val_metric_value,
            self.cumulative_epoch_count,
        )

        for key, value in prepare_for_tensorboard(statistics).items():
            if isinstance(value, (int, float)):
                self.add_tensorboard_scalar(
                    "compression/statistics/{0}".format(key),
                    value,
                    self.cumulative_epoch_count,
                )

        self.dump_checkpoint(model, compression_controller)

    def calculate_minimal_tolerable_accuracy(self, uncompressed_model_accuracy: float) -> None:
        if self.maximal_absolute_accuracy_drop is not None:
            self.minimal_tolerable_accuracy = uncompressed_model_accuracy - self.maximal_absolute_accuracy_drop
        else:
            self.minimal_tolerable_accuracy = uncompressed_model_accuracy * (
                1 - 0.01 * self.maximal_relative_accuracy_drop
            )

    def dump_checkpoint(self, model: TModel, compression_controller: CompressionAlgorithmController) -> None:
        is_best_checkpoint = (
            self.best_val_metric_value == self.current_val_metric_value
            and self.is_model_fully_compressed(compression_controller)
        )
        if not self.dump_checkpoints and not is_best_checkpoint:
            return

        if self._dump_checkpoint_fn is not None:
            checkpoint_path = self._dump_checkpoint_fn(model, compression_controller, self, self._checkpoint_save_dir)
        else:
            checkpoint_path = self._make_checkpoint_path(is_best=False)
            self._save_checkpoint(model, compression_controller, checkpoint_path)
        nncf_logger.info(f"Saved the checkpoint to {checkpoint_path}")

        if is_best_checkpoint:
            self._save_best_checkpoint(model, compression_controller)

    def configure_optimizers(self) -> None:
        self.optimizer, self.lr_scheduler = self._configure_optimizers_fn()

    def initialize_training_loop_fns(
        self,
        train_epoch_fn: Callable[
            [
                CompressionAlgorithmController,
                TModel,
                Optional[int],
                Optional[OptimizerType],
                Optional[LRSchedulerType],
            ],
            float,
        ],
        validate_fn: Callable[[TModel], float],
        configure_optimizers_fn: Optional[Callable[[], Tuple[OptimizerType, LRSchedulerType]]],
        dump_checkpoint_fn: Optional[
            Callable[
                [
                    TModel,
                    CompressionAlgorithmController,
                    "TrainingRunner",
                    Optional[Union[str, pathlib.Path]],
                ],
                None,
            ]
        ],
        load_checkpoint_fn: Optional[Callable[[Checkpoint, Union[str, pathlib.Path]], None]] = None,
        early_stopping_fn: Optional[Callable[[float], bool]] = None,
        update_learning_rate_fn: Optional[Callable[[], None]] = None,
    ) -> None:
        self._train_epoch_fn = train_epoch_fn  # type: ignore
        self._validate_fn = validate_fn  # type: ignore
        self._configure_optimizers_fn = configure_optimizers_fn  # type: ignore
        self._dump_checkpoint_fn = dump_checkpoint_fn  # type: ignore
        self._load_checkpoint_fn = load_checkpoint_fn  # type: ignore
        self._early_stopping_fn = early_stopping_fn
        self._update_learning_rate_fn = update_learning_rate_fn

    def initialize_logging(
        self,
        log_dir: Optional[Union[str, pathlib.Path]] = None,
        tensorboard_writer: Optional[TensorboardWriterType] = None,
    ) -> None:
        self._log_dir = log_dir if log_dir is not None else osp.join(os.getcwd(), "runs")
        self._log_dir = configure_accuracy_aware_paths(self._log_dir)
        self._checkpoint_save_dir = self._log_dir
        self._tensorboard_writer = tensorboard_writer

    def stop_training(self, compression_controller: CompressionAlgorithmController) -> bool:
        if self.is_model_fully_compressed(compression_controller) and self._early_stopping_fn is not None:
            return self._early_stopping_fn(self.current_val_metric_value)
        return False

    def _save_best_checkpoint(self, model: TModel, compression_controller: CompressionAlgorithmController) -> None:
        best_path = self._make_checkpoint_path(is_best=True)
        self._best_checkpoint = (best_path, compression_controller.compression_rate)
        self._save_checkpoint(model, compression_controller, best_path)
        nncf_logger.info(f"Saved the best model to {best_path}")

    def load_best_checkpoint(self, model: TModel) -> float:
        resuming_checkpoint_path: Union[str, pathlib.Path] = ""
        compression_rate: float = 0.0
        resuming_checkpoint_path, compression_rate = self._best_checkpoint  # type: ignore
        nncf_logger.info(f"Loading the best checkpoint found during training: {resuming_checkpoint_path}")
        self._load_checkpoint(model, resuming_checkpoint_path)
        return compression_rate

    def is_model_fully_compressed(self, compression_controller: CompressionAlgorithmController) -> bool:
        return compression_controller.compression_stage() == CompressionStage.FULLY_COMPRESSED

    @abstractmethod
    def add_tensorboard_scalar(self, key: str, data: Union[int, float], step: int) -> None:
        """
        Add a scalar to tensorboard

        :param key: Scalar key name
        :param data: Scalar data
        :param step: Logging step
        """

    @abstractmethod
    def add_tensorboard_image(self, key: str, data: ImageType, step: int) -> None:
        """
        Add an image to tensorboard

        :param key: Image key name
        :param data: Image data
        :param step: Logging step
        """

    @abstractmethod
    def _save_checkpoint(
        self,
        model: TModel,
        compression_controller: CompressionAlgorithmController,
        checkpoint_path: Union[str, pathlib.Path],
    ) -> None:
        """
        Save a model to the disk.

        :param model: The model to be saved
        :param checkpoint_path: The path to save the checkpoint to
        :param compression_controller: The compression controller to be used during
            model fine-tuning
        """

    @abstractmethod
    def _load_checkpoint(self, model: TModel, checkpoint_path: Union[str, pathlib.Path]) -> None:
        """
        Load model from path.

        :param model: The model object in which the state will be loaded.
        :param checkpoint_path: The path where model checkpoint is stored.
        """

    @abstractmethod
    def _make_checkpoint_path(
        self, is_best: bool, compression_rate: Optional[float] = None
    ) -> Union[str, pathlib.Path]:
        """
        Make a path to save the checkpoint there

        :param is_best: Whether this is the best model yet
        :param compression_rate: Compression rate of the model
        """


class BaseAdaptiveCompressionLevelTrainingRunner(BaseAccuracyAwareTrainingRunner):
    """
    The base adaptive compression level accuracy-aware training Runner object,
    initialized with the default parameters unless specified in the config.
    """

    def __init__(
        self,
        accuracy_aware_training_params: Dict[str, Union[float, int]],
        uncompressed_model_accuracy: float,
        verbose: bool = True,
        dump_checkpoints: bool = True,
        lr_updates_needed: bool = True,
        minimal_compression_rate: float = 0.0,
        maximal_compression_rate: float = 0.95,
    ):
        super().__init__(
            accuracy_aware_training_params,
            uncompressed_model_accuracy,
            verbose,
            dump_checkpoints,
            lr_updates_needed,
        )

        self.compression_rate_step: float = accuracy_aware_training_params.get(
            "initial_compression_rate_step", AA_INITIAL_COMPRESSION_RATE_STEP
        )
        self.compression_rate_step_reduction_factor: float = accuracy_aware_training_params.get(
            "compression_rate_step_reduction_factor",
            AA_COMPRESSION_RATE_STEP_REDUCTION_FACTOR,
        )

        self.lr_reduction_factor: float = accuracy_aware_training_params.get(
            "lr_reduction_factor", AA_LR_REDUCTION_FACTOR
        )
        self.minimal_compression_rate_step: float = accuracy_aware_training_params.get(
            "minimal_compression_rate_step", AA_MINIMAL_COMPRESSION_RATE_STEP
        )
        self.patience_epochs = accuracy_aware_training_params.get("patience_epochs", AA_PATIENCE_EPOCHS)
        self.initial_training_phase_epochs = accuracy_aware_training_params.get(
            "initial_training_phase_epochs", AA_INITIAL_TRAINING_PHASE_EPOCHS
        )

        self.minimal_compression_rate = minimal_compression_rate
        self.maximal_compression_rate = maximal_compression_rate

        self._best_checkpoints: Dict[float, Tuple[Union[str, pathlib.Path], float]] = {}
        self._compression_rate_target: Optional[float] = None
        self.adaptive_controller: CompressionAlgorithmController
        self.was_compression_increased_on_prev_step: Optional[float] = None

    def dump_statistics(self, model: TModel, compression_controller: CompressionAlgorithmController) -> None:
        self.update_training_history(self.compression_rate_target, self.current_val_metric_value)
        super().dump_statistics(model, compression_controller)

    def _save_best_checkpoint(self, model: TModel, compression_controller: CompressionAlgorithmController) -> None:
        best_path = self._make_checkpoint_path(is_best=True, compression_rate=self.compression_rate_target)

        accuracy_budget = self.best_val_metric_value - self.minimal_tolerable_accuracy
        if (
            self.compression_rate_target in self._best_checkpoints
            and self._best_checkpoints[self.compression_rate_target][1] >= accuracy_budget
        ):
            return

        self._best_checkpoints[self.compression_rate_target] = (
            best_path,
            accuracy_budget,
        )
        self._save_checkpoint(model, compression_controller, best_path)
        nncf_logger.info(f"Saved the best model to {best_path}")

    def load_best_checkpoint(self, model: TModel) -> float:
        # load checkpoint with the highest compression rate and positive acc budget
        possible_checkpoint_rates = self.get_compression_rates_with_positive_acc_budget()
        if len(possible_checkpoint_rates) == 0:
            nncf_logger.warning(
                "Could not produce a compressed model satisfying the set accuracy "
                "degradation criterion during training. Increasing the number of training "
                "epochs"
            )
            return self.compression_rate_target

        best_checkpoint_compression_rate = None
        for checkpoint_rate in sorted(possible_checkpoint_rates, key=lambda x: -x):
            if checkpoint_rate in self._best_checkpoints:
                best_checkpoint_compression_rate = checkpoint_rate
                break
        if best_checkpoint_compression_rate is None:
            nncf_logger.error(
                "Could not load the model - no models with positive accuracy budget in compression training history."
            )
            return self.compression_rate_target

        resuming_checkpoint_path: Union[str, pathlib.Path] = self._best_checkpoints[best_checkpoint_compression_rate][0]
        nncf_logger.info(f"Loading the best checkpoint found during training: {resuming_checkpoint_path}")
        self._load_checkpoint(model, resuming_checkpoint_path)
        return best_checkpoint_compression_rate

    @property
    def compression_rate_target(self) -> float:
        if self._compression_rate_target is None:
            return self.adaptive_controller.compression_rate
        return self._compression_rate_target

    @compression_rate_target.setter
    def compression_rate_target(self, value: float) -> None:
        self._compression_rate_target = value

    def update_training_history(self, compression_rate: float, metric_value: float) -> None:
        accuracy_budget = metric_value - self.minimal_tolerable_accuracy
        self._compressed_training_history.append((compression_rate, accuracy_budget))

        if IMG_PACKAGES_AVAILABLE:
            with noninteractive_plotting():
                fig = plt.figure()
                plt.plot(
                    cast(List[float], self.compressed_training_history.keys()),
                    cast(List[float], self.compressed_training_history.values()),
                )
                buf = io.BytesIO()
                plt.savefig(buf, format="jpeg")
                buf.seek(0)
                image = PIL.Image.open(buf)
                self.add_tensorboard_image(
                    "compression/accuracy_aware/acc_budget_vs_comp_rate",
                    image,
                    len(self.compressed_training_history),
                )
                plt.close(fig)

    @property
    def compressed_training_history(self) -> Dict[float, float]:
        return dict(self._compressed_training_history)

    def get_compression_rates_with_positive_acc_budget(self) -> List[float]:
        return [comp_rate for (comp_rate, acc_budget) in self._compressed_training_history if acc_budget >= 0]



# Copyright (c) 2024 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from enum import Enum
from typing import List, Optional, Set, Tuple

from nncf.common.quantization.structs import QuantizerConfig
from nncf.common.quantization.structs import UnifiedScaleType


class QuantizationTrait(Enum):
    """
    General, hardware-agnostic specifications for the relation of operators to quantization.
    Hardware-specific quantization configuration is handled elsewhere.
    """

    NON_QUANTIZABLE = -1
    QUANTIZATION_AGNOSTIC = 0
    INPUTS_QUANTIZABLE = 1

    # For embeddings, the inputs are integer and are used to index into the weight of the layer,
    # therefore if the weight is already quantized there may be no need to quantize the outgoing tensor.
    # TODO: unify scales for such operations if they become linked through a downstream op (such as
    # two embeddings added to each other)
    OUTPUT_QUANTIZATION_AS_WEIGHTS = 2

    # A concat node should ultimately either have all of its inputs quantized with the same configuration
    # or none at all. This cannot be determined ahead of time, hence the special trait.
    CONCAT = 100


class PropagatingQuantizer:
    """
    Used in conjunction with QuantizerPropagationStateGraph to keep track of
    the allowed quantization configs corresponding to the model operation node
    whose inputs it quantizes, and also of the nodes/edges in the model control
    graph that this quantizer affects. It should be moved against the data flow of
    the model, tracking the affected nodes and edges of
    QuantizerPropagationStateGraph. No actual quantization modules are used here,
    only the associated configs (such as bitwidths, modes, signed/unsigned
    attributes etc.)
    """

    def __init__(
        self,
        id_: int,
        quant_configs: List[QuantizerConfig],
        init_location_node_key: str,
        unified_scale_type: Optional[UnifiedScaleType] = None,
    ):
        """
        :param id_: The unique identifier of the new propagating quantizer.
        :param quant_configs: The quantizer configurations that this quantizer currently allows.
        :param init_location_node_key: The node key in QuantizerPropagationStateGraph that this
          quantizer is being inserted into.
        :param unified_scale_type: The type of unified scales for this quantizer - if unspecified,
          this quantizer won't require unified scales.
        """
        self.potential_quant_configs: List[QuantizerConfig] = quant_configs
        self.affected_edges = set()
        self.affected_ip_nodes: Set[str] = set()
        self.propagation_path: PropagationPath = []
        self.current_location_node_key = init_location_node_key
        self.last_accepting_location_node_key = None
        self.id = id_
        self.unified_scale_type = unified_scale_type
        self.affected_operator_nodes = set()
        self.quantized_input_sink_operator_nodes = set()
        self.downstream_propagating_quantizers = set()

    def __eq__(self, other):
        return self.id == other.id

    def __hash__(self):
        return hash(self.id)


class QuantizerPropagationStateGraphNodeType(Enum):
    PRE_HOOK = 0
    POST_HOOK = 1
    OPERATOR = 2
    AUXILIARY_BARRIER = 3


class SharedAffectedOpsPropagatingQuantizerGroup:
    """Combines propagating quantizers that share affected operations"""

    def __init__(self, affecting_prop_quants: Set[PropagatingQuantizer], affected_op_node_keys: Set[str]):
        self.affecting_prop_quants: Set[PropagatingQuantizer] = affecting_prop_quants
        self.affected_op_node_keys: Set[str] = affected_op_node_keys

    def update(self, other: "SharedAffectedOpsPropagatingQuantizerGroup"):
        self.affected_op_node_keys.update(other.affected_op_node_keys)
        self.affecting_prop_quants.update(other.affecting_prop_quants)


class IgnoreReason(Enum):
    USER_REQUESTED = 0
    AUTOGENERATED = 1


PropagationPath = List[Tuple[str, str]]



# Copyright (c) 2024 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import math
from abc import abstractmethod
from typing import Any, Dict, List, Optional, TypeVar

from nncf.common.schedulers import BaseCompressionScheduler
from nncf.experimental.torch.nas.bootstrapNAS.training.stage_descriptor import StageDescriptor

OptimizerType = TypeVar("OptimizerType")


def adjust_learning_rate(
    optimizer: OptimizerType,
    epoch: float,
    init_lr: float,
    epochs: float,
    batch: float = 0,
    n_batch: float = 0,
    lr_schedule_type: str = "cosine",
):
    new_lr = calc_learning_rate(epoch, init_lr, epochs, batch, n_batch, lr_schedule_type)
    for param_group in optimizer.param_groups:
        param_group["lr"] = new_lr
    return new_lr


def warmup_adjust_learning_rate(
    optimizer: OptimizerType,
    init_lr: float,
    t_total: float,
    n_batch: float,
    epoch: float,
    batch: float = 0,
    warmup_lr: float = 0,
):
    t_cur = epoch * n_batch + batch + 1
    new_lr = t_cur / t_total * (init_lr - warmup_lr) + warmup_lr
    for param_group in optimizer.param_groups:
        param_group["lr"] = new_lr
    return new_lr


def calc_learning_rate(
    epoch: float,
    init_lr: float,
    n_epochs: float,
    batch: float = 0,
    n_batch: float = 0,
    lr_schedule_type: str = "cosine",
):
    if lr_schedule_type == "cosine":
        t_total = n_epochs * n_batch
        t_cur = epoch * n_batch + batch
        lr = 0.5 * init_lr * (1 + math.cos(math.pi * t_cur / t_total))
    elif lr_schedule_type is None:
        lr = init_lr
    else:
        raise ValueError("do not support: %s" % lr_schedule_type)
    return lr


class LRSchedulerParams:
    """
    Storage class for LR Scheduler parameters
    """

    def __init__(
        self,
        num_steps_in_epoch: float,
        base_lr: float = 3.4e-4,
        num_epochs: float = 0,
        warmup_epochs: float = 0,
        warmup_lr: float = 0,
    ):
        """
        Initializes storage class for learning rate scheduler parameters

        :param num_steps_in_epoch:
        :param base_lr:
        :param num_epochs:
        :param warmup_epochs:
        :param warmup_lr:
        """
        self.num_steps_in_epoch = num_steps_in_epoch
        self.base_lr = base_lr
        self.num_epochs = num_epochs
        self.warmup_epochs = warmup_epochs
        self.warmup_lr = warmup_lr

    @classmethod
    def from_dict(cls, lr_scheduler_config: Dict[str, Any]) -> "LRSchedulerParams":
        """
        Initialize learning rate scheduler parameters storage clas from Dict.
        :param lr_scheduler_config: Dict with parameters of learning rate scheduler.
        :return:
        """
        num_steps_in_epoch = lr_scheduler_config.get("num_steps_in_epoch", 0)
        base_lr = lr_scheduler_config.get("base_lr", 3.4e-4)
        num_epochs = lr_scheduler_config.get("num_epochs", 0)
        warmup_epochs = lr_scheduler_config.get("warmup_epochs", 0)
        warmup_lr = lr_scheduler_config.get("warmup_lr", 3.4e-4)
        return cls(num_steps_in_epoch, base_lr, num_epochs, warmup_epochs, warmup_lr)


class BaseLRScheduler(BaseCompressionScheduler):
    """
    Base class for the learning rate scheduler
    """

    def __init__(self, optimizer: OptimizerType, num_steps_in_epoch: float):
        super().__init__()
        self._optimizer = optimizer
        self._num_steps_in_epoch = num_steps_in_epoch

    @abstractmethod
    def stage_step(self, stage_desc: StageDescriptor):
        pass

    @classmethod
    def from_state(cls, state: Dict[str, Any], optimizer: OptimizerType):
        return cls(optimizer, **state)

    def get_last_lr(self) -> List[Any]:
        return [group["lr"] for group in self._optimizer.param_groups]


class GlobalLRScheduler(BaseLRScheduler):
    """
    Global LR scheduler prevents LR adjustments per stage.
    """

    def __init__(
        self,
        optimizer: OptimizerType,
        num_steps_in_epoch: float,
        *,
        base_lr: float,
        num_epochs: float,
        warmup_epochs: float = 0,
        warmup_lr: float = 3.4e-4
    ):
        super().__init__(optimizer, num_steps_in_epoch)
        self._base_lr = base_lr
        self._num_epochs = num_epochs
        self._warmup_epochs = warmup_epochs
        self._warmup_lr = warmup_lr

    @classmethod
    def from_config(cls, optimizer: OptimizerType, params: "LRSchedulerParams"):
        return cls(optimizer, **params.__dict__)

    def stage_step(self, stage_desc: StageDescriptor):
        # do nothing
        pass

    def step(self, next_step: Optional[int] = None) -> None:
        super().step(next_step)
        step_from_epoch_start = self.current_step - (self.current_epoch * (self._num_steps_in_epoch + 1))
        if self.current_epoch < self._warmup_epochs and self.current_epoch != -1:
            warmup_adjust_learning_rate(
                optimizer=self._optimizer,
                init_lr=self._base_lr,
                t_total=self._warmup_epochs * self._num_steps_in_epoch,
                n_batch=self._num_steps_in_epoch,
                epoch=self.current_epoch,
                batch=step_from_epoch_start,
                warmup_lr=self._warmup_lr,
            )
        else:
            adjust_learning_rate(
                optimizer=self._optimizer,
                epoch=self.current_epoch - self._warmup_epochs,
                init_lr=self._base_lr,
                epochs=self._num_epochs,
                batch=step_from_epoch_start,
                n_batch=self._num_steps_in_epoch,
                lr_schedule_type="cosine",
            )

    def get_state(self) -> Dict[str, Any]:
        state_dict = {
            "num_steps_in_epoch": self._num_steps_in_epoch,
            "base_lr": self._base_lr,
            "num_epochs": self._num_epochs,
            "warmup_epochs": self._warmup_epochs,
            "warmup_lr": self._warmup_lr,
        }
        return state_dict


class StageLRScheduler(BaseLRScheduler):
    """
    Stage learning rate scheduler. Allows adjustment of the learning rate at a stage transition.
    """

    def __init__(self, optimizer: OptimizerType, num_steps_in_epoch: float):
        super().__init__(optimizer, num_steps_in_epoch)
        self._init_lr = None
        self._num_epochs = None

    @classmethod
    def from_config(cls, optimizer: OptimizerType, params: LRSchedulerParams):
        return cls(optimizer, params.num_steps_in_epoch)

    def stage_step(self, stage_desc: StageDescriptor):
        self.reset(stage_desc.init_lr, stage_desc.epochs_lr)

    def step(self, next_step: Optional[int] = None) -> None:
        super().step(next_step)
        step_from_epoch_start = self.current_step - (self.current_epoch * (self._num_steps_in_epoch + 1))
        adjust_learning_rate(
            optimizer=self._optimizer,
            epoch=self.current_epoch,
            init_lr=self._init_lr,
            epochs=self._num_epochs,
            batch=step_from_epoch_start,
            n_batch=self._num_steps_in_epoch,
            lr_schedule_type="cosine",
        )

    def get_state(self) -> Dict[str, Any]:
        state_dict = {
            "num_steps_in_epoch": self._num_steps_in_epoch,
            "init_lr": self._init_lr,
            "num_epochs": self._num_epochs,
        }
        return state_dict

    def reset(self, init_lr: float, num_epochs: float):
        """
        Resets the learning rate for the current stage

        :param init_lr:
        :param num_epochs:
        :return:
        """
        self._num_epochs = num_epochs
        self._init_lr = init_lr
        self.epoch_step(0)
        self.step(0)



# Copyright (c) 2024 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from collections import namedtuple
from typing import NamedTuple

import networkx as nx
import torch

import nncf
from nncf.common.graph import NNCFNodeName
from nncf.common.quantization.structs import QuantizationScheme as QuantizationMode
from nncf.torch.layers import NNCFConv2d
from nncf.torch.module_operations import UpdatePaddingValue
from nncf.torch.nncf_network import NNCFNetwork
from nncf.torch.quantization.layers import BaseQuantizer
from nncf.torch.quantization.layers import QuantizerConfig
from nncf.torch.quantization.layers import SymmetricQuantizer


class AdjustPaddingArgs(NamedTuple):
    weight_bitwidth: int
    activation_quantizer: BaseQuantizer
    module_op_node_name: NNCFNodeName


class CalculatePaddingAdjustment:
    """
    Calculates padding value to perform a workaround for U4 support on NPU.
    NPU supports only i4 for weights and activations with zero-point=0 and padding=0. This imposes some limitations on
    the quantization scheme we can apply. In case of unsigned input for a quantizer (e.g. output of ReLU) half of
    i4 range (8 values) is insufficient to preserve the accuracy. To overcome the problem it is proposed
    to transform u4 to i4 in the NPU plugin by shifting the input by half of the quantization range to left. Padding
    value should be shifted as well. And to make it zero after the shift (non-zero padding values are not
    supported), the model should be trained with padding value equal to the half of the quantization range.
    """

    def __init__(self, activation_quantizer: SymmetricQuantizer):
        if not isinstance(activation_quantizer, SymmetricQuantizer):
            raise nncf.InternalError("Padding adjustment is not supported for not symmetric quantization")
        self._activation_quantizer = activation_quantizer
        self._is_enabled = True

    def __call__(self, previous_padding_value) -> torch.Tensor:
        if self._is_enabled:
            scale = self._activation_quantizer.scale
            eps = self._activation_quantizer.eps
            safe_scale = abs(scale) + eps
            return safe_scale / 2
        return previous_padding_value

    @staticmethod
    def is_config_applicable(qconfig: QuantizerConfig):
        return (
            not qconfig.per_channel
            and qconfig.num_bits == 4
            and not qconfig.signedness_to_force
            and qconfig.mode == QuantizationMode.SYMMETRIC
        )


def add_adjust_padding_nodes(bitwidth_graph: nx.DiGraph, model: NNCFNetwork) -> nx.DiGraph():
    NewNodeArgs = namedtuple("NewNodeArgs", ("node_key", "attr", "parent_node_key"))
    nncf_graph = model.nncf.get_graph()
    args = []
    for node_key in bitwidth_graph.nodes:
        node = nncf_graph.get_node_by_key(node_key)
        module = model.nncf.get_containing_module(node.node_name)
        if isinstance(module, NNCFConv2d):
            adjust_padding_ops = filter(lambda x: isinstance(x, UpdatePaddingValue), module.pre_ops.values())
            for _ in adjust_padding_ops:
                new_node_key = f"{node_key}_apad"
                attr = dict(type="", label="adjust_padding_value", style="filled", color="yellow")
                args.append(NewNodeArgs(new_node_key, attr, node_key))

    for arg in args:
        bitwidth_graph.add_node(arg.node_key, **arg.attr)
        bitwidth_graph.add_edge(arg.node_key, arg.parent_node_key)
    return bitwidth_graph



# Copyright (c) 2024 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from dataclasses import dataclass
from typing import Optional, Tuple

import numpy as np

from nncf.experimental.tensor import functions as fns
from nncf.quantization.fake_quantize import FakeQuantizeParameters
from nncf.quantization.fake_quantize import calculate_scale_zero_point


@dataclass
class ONNXQuantizerLayerParameters:
    """
    Class handles Quantizer layer attributes.

    :param scale: Quantizer scale.
    :param zero_point: Quantizer zero point.
    :param tensor_type: Signed or Unsigned tensor type.
    :param axis: Axis for per-channel quantization. Should be none in case of per-tensor.
    """

    scale: np.ndarray
    zero_point: np.ndarray
    tensor_type: np.dtype
    axis: Optional[int] = None


def convert_fq_params_to_onnx_params(
    parameters: FakeQuantizeParameters, num_bits: int, tensor_type: np.dtype, axis: Tuple[int]
) -> ONNXQuantizerLayerParameters:
    """
    Converts common FakeQuantizeParameters to ONNXQuantizerLayerParameters.

    :param parameters: FakeQuantizeParameters representation.
    :param num_bits: Number of quantizer bits.
    :param tensor_type: Value type of the tensor. Could be INT8 or UINT8.
    :param axis: Axis for per-channel quantization.
    :return: Quantizer layer attributes.
    """
    if num_bits != 8:
        raise ValueError("Can only export to INT8/UIN8 8 bits ONNX Quantize/Dequantize pairs.")

    levels = parameters.levels
    if levels not in [255, 256]:
        raise ValueError("Can only export to INT8/UIN8 256-level ONNX Quantize/Dequantize pairs.")

    input_low, input_high = parameters.input_low, parameters.input_high
    output_low, output_high = parameters.output_low, parameters.output_high
    if not fns.allclose(input_high, output_high) or not fns.allclose(input_low, output_low):
        raise ValueError(
            "ONNX Quantize/Dequantize pairs only support input_high == output_high and input_low == output_low."
        )

    level_low, level_high = get_level_low_level_high(tensor_type)
    narrow_range = levels == 2**num_bits - 1
    scale, zero_point = calculate_scale_zero_point(input_low, input_high, level_low, level_high, narrow_range)
    # ONNX demands parameters to be a scalar or 1-D Tensor.
    scale = np.squeeze(scale)
    zero_point = np.squeeze(zero_point)
    # ONNX axis parameter format specification.
    axis = None if not axis else axis[0]
    return ONNXQuantizerLayerParameters(scale.data, zero_point.data, tensor_type, axis)


def get_level_low_level_high(tensor_type: np.dtype) -> Tuple[int, int]:
    """
    Returns the minimum and maximum level for the quantizer.
    In ONNX opset Q/DequantizeLinear-13 uses only two levels: [-128, 127] and [0, 255].

    :param tensor_type: Value type of the tensor. Could be INT8 or UINT8.
    :return: Minimum level and maximum level of the quantizer.
    """
    return (0, 255) if tensor_type == np.uint8 else (-128, 127)



# Copyright (c) 2024 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from copy import deepcopy
from dataclasses import dataclass
from typing import Any, Dict, List, Optional, TypeVar

from nncf import Dataset
from nncf import nncf_logger
from nncf.common.graph.graph import NNCFGraph
from nncf.common.graph.graph import NNCFNode
from nncf.common.graph.graph_matching import find_subgraphs_matching_pattern
from nncf.common.logging.track_progress import track
from nncf.common.tensor_statistics.statistic_point import StatisticPointsContainer
from nncf.common.utils.backend import BackendType
from nncf.common.utils.backend import get_backend
from nncf.experimental.tensor import functions as fns
from nncf.quantization.algorithms.algorithm import Algorithm
from nncf.quantization.algorithms.weight_compression.config import WeightCompressionParameters
from nncf.quantization.algorithms.weight_compression.weight_lowering import do_dequantization
from nncf.quantization.algorithms.weight_compression.weight_lowering import do_integer_quantization
from nncf.quantization.passes import transform_to_inference_graph

TModel = TypeVar("TModel")
TTensor = TypeVar("TTensor")
TWeightType = TypeVar("TWeightType")


@dataclass
class AWQCompressionInfo:
    """
    Information on AWQ nodes.
    """

    weight_params: WeightCompressionParameters = None
    target_node: NNCFNode = None
    merge_node: NNCFNode = None


class AWQ(Algorithm):
    """
    Modified AWQ algorithm implementation.
    """

    def __init__(
        self,
        model: TModel,
        name_to_node_mapping: Dict[str, Any],
        all_weight_params: List[WeightCompressionParameters],
        nodes_to_compress: List[NNCFNode],
        activations: Optional[Dict[str, TTensor]] = None,
        subset_size: int = 32,
        percent_to_apply=0.002,
        alpha_min=0.0,
        alpha_max=1.0,
        steps=100,
    ):
        """
        :param model: Model for applying algorithm.
        :param name_to_node_mapping: Name to node mapping for updating node weights.
        :param all_weight_params: List of all weight parameters.
        :param nodes_to_compress: List of nodes for processing.
        :param activations: The input activations of the layers considered for compression.
        :param subset_size: The number of samples for AWQ.
        :param percent_to_apply: The percent of outliers for correction.
        :param alpha_min: Minimum value of smoothness parameter for grid search.
        :param alpha_max: Maximal value of smoothness parameter for grid search.
        :param steps: The number of the steps in grid search.
        """
        super().__init__()
        self.name_to_node_mapping = name_to_node_mapping
        self._all_weight_params = all_weight_params
        self._nodes_to_compress = nodes_to_compress
        self._activations = activations
        self._subset_size = subset_size
        self._percent_to_apply = percent_to_apply
        self._alpha_min = alpha_min
        self._alpha_max = alpha_max
        self._steps = steps
        self._backend_entity = None
        self._patterns = None

        self._set_backend_entity(model)

    @property
    def available_backends(self) -> List[BackendType]:
        return [BackendType.OPENVINO]

    def _set_backend_entity(self, model: TModel) -> None:
        """
        Creates a helper class with a backed-specific logic of the algorithm.

        :param model: Backend-specific input model.
        :param all_weight_params: List of all weight parameters.
        :param nodes_to_compress: List of nodes for processing.
        :param activations: The input activations of the layers considered for compression.
        """

        model_backend = get_backend(model)
        if model_backend == BackendType.OPENVINO:
            from nncf.quantization.algorithms.weight_compression.openvino_backend import OVAWQAlgoAlgoBackend

            self._backend_entity = OVAWQAlgoAlgoBackend(model, self.name_to_node_mapping)
            self._patterns = self._backend_entity.get_awq_patterns()
        else:
            raise RuntimeError(
                "Cannot return backend-specific AWQ entity because {} is not supported!".format(model_backend.value)
            )

    def apply(
        self,
        model: TModel,
        graph: NNCFGraph,
        statistic_points: Optional[StatisticPointsContainer] = None,
        dataset: Optional[Dataset] = None,
    ) -> TModel:
        """
        Applies the algorithm to the model.

        :param model: Model for applying algorithm.
        :param graph: Model graph.
        :param statistic_points: Statistic points with collected statistics values.
        :param dataset: A representative dataset for the calibration process.
        :return: A resulting model.
        """
        matches = []

        inference_nncf_graph = transform_to_inference_graph(deepcopy(graph), [], [], [])
        nx_graph = inference_nncf_graph.get_nx_graph_copy()
        for _, pattern_graph in self._patterns.items():
            matches.extend(find_subgraphs_matching_pattern(nx_graph, pattern_graph(), strict=False))

        if len(matches) == 0:
            nncf_logger.info("No matching patterns were found for applying AWQ algorithm, it will be skipped.")
            return model

        target_node_names = []
        merge_node_names = []
        awq_data = {}
        name_mapping = {wp.weight_name: idx for idx, wp in enumerate(self._all_weight_params)}

        for match in matches:
            nncf_node = graph.get_node_by_key(match[-1])
            if not self._backend_entity.is_node_with_weights(nncf_node, graph):
                continue

            for weight_op_friendly_name, _ in self._backend_entity.get_weight_names_and_port_ids(nncf_node, graph):
                target_node_names.append(weight_op_friendly_name)

            nncf_node = graph.get_node_by_key(match[0])
            for weight_op_friendly_name, _ in self._backend_entity.get_weight_names_and_port_ids(nncf_node, graph):
                merge_node_names.append(weight_op_friendly_name)

            assert len(target_node_names) == len(merge_node_names)
            weight_params = self._all_weight_params[name_mapping[target_node_names[-1]]]
            if weight_params.compression_config.num_bits != 4:
                continue
            target_node = self._nodes_to_compress[name_mapping[target_node_names[-1]]]
            merge_node = self._nodes_to_compress[name_mapping[merge_node_names[-1]]]

            awq_data[target_node.node_name] = AWQCompressionInfo(weight_params, target_node, merge_node)

        alpha_step = (self._alpha_max - self._alpha_min) / self._steps

        for k, awq_data_item in track(awq_data.items(), description="Applying AWQ"):
            wp = awq_data_item.weight_params
            target_node = awq_data_item.target_node
            merge_node = awq_data_item.merge_node

            weight_data = self._backend_entity.get_weight_names_and_port_ids(wp.node_with_weight, graph)
            if len(weight_data) != 1:  # not supported by the algorithm
                continue
            _, weight_port_id = weight_data[0]

            config = wp.compression_config

            stats = self._activations[k]
            X = fns.stack([fns.mean(stat, axis=0) for stat in stats])
            X = fns.transpose(X)

            s = fns.max(fns.abs(X), axis=1)

            if X.shape[1] > self._subset_size:
                lens = [stat.shape[0] for stat in stats]
                step = X.shape[1] // self._subset_size
                idxs = [i[0] for i in sorted(enumerate(lens), key=lambda x: -x[1])][::step]
                X = X[:, idxs]

            top_k = max(int(s.shape[0] * self._percent_to_apply), 1)
            topk_idxs = fns.argsort(-s)[:top_k]

            group_size = config.group_size
            if group_size == -1:
                group_size = s.shape[0]

            groups_to_correct = set()
            for idx in topk_idxs:
                groups_to_correct.add(idx.data // group_size)

            groups_to_correct = list(groups_to_correct)

            weight = self._backend_entity.get_weight(
                wp.node_with_weight, weight_port_id, model, graph
            )  # get_const_value(wp.weight_node)
            assert isinstance(wp.reduction_axes, tuple) and len(wp.reduction_axes) == 1
            reduction_axis = wp.reduction_axes[0]

            if reduction_axis == 0:
                weight = fns.transpose(weight)
                reduction_axis = 1

            shape_vector = fns.mean(X, axis=1)
            scale = fns.ones_like(shape_vector)

            awq_config = deepcopy(config)
            awq_config.group_size = -1

            for gi in groups_to_correct:
                offset = gi * group_size
                gscale = s[offset : offset + group_size]

                a_min = fns.quantile(gscale, 0.1)
                a_max = 1e2
                gscale = fns.clip(gscale, a_min=a_min, a_max=a_max)

                gweight = weight[:, offset : offset + group_size]
                gacts = X[offset : offset + group_size, :]

                fp32_out = fns.matmul(gweight, gacts)
                min_diff = fns.max(fns.abs(fp32_out))
                best_scale = None

                alpha = self._alpha_min
                for _ in range(self._steps):
                    cur_scale = gscale**alpha

                    g_compressed_weighs, g_c_scale, g_c_zp = do_integer_quantization(
                        gweight * cur_scale, reduction_axis, awq_config
                    )
                    g_decompressed_weighs = do_dequantization(g_compressed_weighs, g_c_scale, g_c_zp)
                    sacts = gacts / fns.unsqueeze(cur_scale, 1)

                    cur_out = fns.matmul(g_decompressed_weighs, sacts)
                    cur_diff = fns.mean(fns.abs(cur_out - fp32_out))
                    if cur_diff < min_diff:
                        min_diff = cur_diff
                        best_scale = cur_scale
                    alpha += alpha_step

                if best_scale is not None:
                    scale.data[offset : offset + group_size] = best_scale.data

            a_scale = scale
            w_scale = scale
            if wp.reduction_axes[0] == 0:
                w_scale = fns.unsqueeze(w_scale, 1)
                a_scale = fns.unsqueeze(1.0 / a_scale, 0)
            else:
                w_scale = fns.unsqueeze(w_scale, 0)
                a_scale = fns.unsqueeze(1.0 / a_scale, 1)

            scaled_weight = weight * w_scale
            self._backend_entity.set_weight(wp.node_with_weight, weight_port_id, model, graph, scaled_weight)

            for _, port_id in self._backend_entity.get_weight_names_and_port_ids(merge_node, graph):
                merge_weight = self._backend_entity.get_weight(merge_node, port_id, model, graph)
                merge_weight = merge_weight * a_scale
                self._backend_entity.set_weight(merge_node, port_id, model, graph, merge_weight)

            # update activations for next usage
            a_scale_t = fns.transpose(a_scale)
            for i, stat in enumerate(self._activations[k]):
                stat = stat * a_scale_t
                self._activations[k][i] = stat

        return model

    def get_statistic_points(self, model: TModel, graph: NNCFGraph) -> StatisticPointsContainer:
        """
        Returns statistic points, for which StatisticsCollector should collect statistics.

        :param model: Model for statistics collection.
        :param graph: Model graph.
        :return: Statistic points, for which StatisticsCollector should collect statistics.
        """
        return StatisticPointsContainer()
